/*********************************************************************************
 *
 *  The C++ binding for LDB
 *  Copyright (c) 2003 by Leonidas Fegaras, the University of Texas at
 *     Arlington. All rights reserved.
 *  Programmer: Leonidas Fegaras
 *  Date: 02/07/03
 *
 ********************************************************************************/


#include <stdio.h>
#include <errno.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <stream.h>
#include <stdarg.h>
#include "client.h"


const size_t port_number = 9737;


/* RPC tags for remote procedures */
short n = 0;
const short Cflush = n++;
const short CgetInt = n++;
const short CgetString = n++;
const short CgetFloat = n++;
const short Csize = n++;
const short Cnth = n++;
const short CtoString = n++;
const short CgetType = n++;
const short Copen = n++;
const short Cclose = n++;
const short Cquery = n++;
const short Cupdate = n++;
const short Cinsert = n++;
const short Cremove = n++;
const short Ccreatei = n++;
const short Cdropi = n++;
const short Cchange = n++;
const short Ctclschema = n++;
const short Cbegin = n++;
const short Ccommit = n++;
const short Cabort = n++;
const short Cprocess = n++;
const short Cstmt = n++;


class RPCstream {
public:
  int sockfd;
  bool opened;
  RPCstream ( int fd ): sockfd(fd), opened(false) {};
  RPCstream (): sockfd(0), opened(false) {};
  RPCstream& rpc_send ( void* data, size_t size );
  RPCstream& rpc_receive ( void* &data, size_t size );
  RPCstream& rpc_sendv ( void* data, size_t size );
  RPCstream& rpc_receivev ( void* &data, size_t size );
  void close ();
  ~RPCstream () { close(); };
  void send ( String v );
  void send ( int n );
  void send ( short n );
  void send ( float n );
  String receiveString ();
  int receiveInt ();
  short receiveShort ();
  float receiveFloat ();
  void rpc_validate ();
};


const char* cform ( const char* fmt, ... ) {
  int size = 100;
  char* p;
  va_list ap;
  if ((p = new char[size]) == NULL)
     return NULL;
  while (true) {
    va_start(ap, fmt);
    int n = vsnprintf(p, size, fmt, ap);
    va_end(ap);
    if (n > -1 && n < size)
      return p;
    if (n > -1)
       size = n+1;
    else size *= 2;
    delete p;
    if ((p = new char[size]) == NULL)
       return NULL;
  };
};


static int client_socket;

static RPCstream rpc_stream;


static void rpc_error ( const char* msg ) {
  cerr << "RPC Error: " << msg << ".\n";
  exit(1);
};


RPCstream& RPCstream::rpc_send ( void* data, size_t size ) {
  if (::send(sockfd,(char*)data,size,0) == size)
     return *this;
  else rpc_error("Incomplete data in socket");
};


RPCstream& RPCstream::rpc_receive ( void* &data, size_t size ) {
  if (recv(sockfd,(char*)data,size,MSG_WAITALL) == size)
     return *this;
  else rpc_error("Incomplete data from socket");
};


#if defined(linux) || defined(__linux__)
// Need to convert big-endian data (low byte first) into little-endian
// (high byte first), since Java's DataStreams are little-endian
RPCstream& RPCstream::rpc_sendv ( void* data, size_t size ) {
  for(int i=size-1; i>=0; i--)
     rpc_send((void*)(((char*)data)+i),1); 
  return *this;
};
RPCstream& RPCstream::rpc_receivev ( void* &data, size_t size ) {
  for(int i=size-1; i>=0; i--)
  {  void* d = (void*)(((char*)data)+i);
     rpc_receive(d,1);
  };
  return *this;
};
#else
// little-endian (eg, DEC)
RPCstream& RPCstream::rpc_sendv ( void* data, size_t size ) {
  return rpc_send(data,size);
};
RPCstream& RPCstream::rpc_receivev ( void* &data, size_t size ) {
   return rpc_receive(data,size);
};
#endif


void RPCstream::close () {
  if (opened)
  {  rpc_stream.send((short) -1);
     ::close(client_socket);
  };
};


void RPCstream::send ( String s ) {
  int len = s.length();
  send(len);
  rpc_send((void*) s.data(),len);
};


void RPCstream::send ( int n ) {
  rpc_sendv(&n,sizeof(int));
};


void RPCstream::send ( short n ) {
  rpc_sendv(&n,sizeof(short));
};


void RPCstream::send ( float n ) {
  rpc_sendv(&n,sizeof(float));
};


String RPCstream::receiveString () {
  int length = receiveInt();
  const char* content = new char[length];
  rpc_receive((void*) content,length);
  return String(content,length);
};


int RPCstream::receiveInt () {
  int* n = new int;
  rpc_receivev((void*) n,sizeof(int));
  return *n;
};


short RPCstream::receiveShort () {
  short* n = new short;
  rpc_receivev((void*) n,sizeof(short));
  return *n;
};


float RPCstream::receiveFloat () {
  float* n = new float;
  rpc_receivev((void*) n,sizeof(float));
  return *n;
};


void RPCstream::rpc_validate () {
  char res;
  recv(sockfd,&res,1,MSG_WAITALL);
  if (res!=0)
     throw ServerError(receiveString());
};


LDBobject::LDBobject () {
  ID = rpc_stream.receiveInt();
};


void LDBobject::operation ( short opr ) const {
  rpc_stream.send(opr);
  rpc_stream.send(ID);
};


void LDBobject::flush () {
  operation(Cflush);
  rpc_stream.rpc_validate();
};


int LDBobject::getInt () const {
  operation(CgetInt);
  rpc_stream.rpc_validate();
  return rpc_stream.receiveInt();
};


String LDBobject::getString () const {
  operation(CgetString);
  rpc_stream.rpc_validate();
  return rpc_stream.receiveString();
};


float LDBobject::getFloat () const {
  operation(CgetFloat);
  rpc_stream.rpc_validate();
  return rpc_stream.receiveFloat();
};


String LDBobject::toString () const {
  operation(CtoString);
  rpc_stream.rpc_validate();
  return rpc_stream.receiveString();
};


String LDBobject::getType () const {
  operation(CgetType);
  rpc_stream.rpc_validate();
  return rpc_stream.receiveString();
};


int LDBobject::size () const {
  operation(Csize);
  rpc_stream.rpc_validate();
  return rpc_stream.receiveInt();
};


LDBobject LDBobject::nth ( int index ) const {
  operation(Cnth);
  rpc_stream.send(index);
  rpc_stream.rpc_validate();
  return LDBobject();
};


void init_sockaddr ( struct sockaddr_in *name, const char *hostname, uint16_t port ) {
  struct hostent *hostinfo;
  name->sin_family = AF_INET;
  name->sin_port = htons(port);
  hostinfo = gethostbyname(hostname);
  if (hostinfo == NULL)
     rpc_error(cform("Unknown host %s.\n", hostname));
  name->sin_addr = *(struct in_addr *) hostinfo->h_addr;
};


void LDB::open ( String host, String username, String password, String schema ) {
  struct sockaddr_in servername;
  client_socket = socket(PF_INET,SOCK_STREAM,0);
  if (client_socket < 0)
     rpc_error("Cannot create a socket for client");
  init_sockaddr(&servername,host.data(),port_number);
  if (connect(client_socket,(struct sockaddr *) &servername,sizeof(servername)) < 0)
     rpc_error("Cannot connect to the server");
  rpc_stream = RPCstream(client_socket);
  rpc_stream.opened = true;
  rpc_stream.send(Copen);
  rpc_stream.send(username);
  rpc_stream.send(password);
  rpc_stream.send(schema);
  rpc_stream.rpc_validate();
  opened = true;
};


void LDB::close () {
  if (opened) {
    if (in_transaction)
       abort_transaction();
    rpc_stream.send(Cclose);
    rpc_stream.rpc_validate();
    opened = false;
  };
};


LDBobject LDB::query ( String query ) {
  rpc_stream.send(Cquery);
  rpc_stream.send(query);
  rpc_stream.rpc_validate();
  return LDBobject();
};


void LDB::statement ( String stmt ) {
  rpc_stream.send(Cstmt);
  rpc_stream.send(stmt);
  rpc_stream.rpc_validate();
};


void LDB::update ( String varname, String query ) {
  rpc_stream.send(Cupdate);
  rpc_stream.send(varname);
  rpc_stream.send(query);
  rpc_stream.rpc_validate();
};


void LDB::insert ( String varname, String query ) {
  rpc_stream.send(Cinsert);
  rpc_stream.send(varname);
  rpc_stream.send(query);
  rpc_stream.rpc_validate();
};


void LDB::remove ( String varname, String query ) {
  rpc_stream.send(Cremove);
  rpc_stream.send(varname);
  rpc_stream.send(query);
  rpc_stream.rpc_validate();
};


void create_index ( String index_name, String class_name, String attribute ) {
  rpc_stream.send(Ccreatei);
  rpc_stream.send(index_name);
  rpc_stream.send(class_name);
  rpc_stream.send(attribute);
  rpc_stream.rpc_validate();
};


void remove_index ( String index_name ) {
  rpc_stream.send(Cdropi);
  rpc_stream.send(index_name);
  rpc_stream.rpc_validate();
};


void LDB::change_password ( String oldp, String newp ) {
  rpc_stream.send(Cchange);
  rpc_stream.send(oldp);
  rpc_stream.send(newp);
  rpc_stream.rpc_validate();
};

String LDB::tcl_schema () {
  rpc_stream.send(Ctclschema);
  rpc_stream.rpc_validate();
  return rpc_stream.receiveString();  
};


void LDB::begin_transaction () {
  if (!in_transaction) {
    rpc_stream.send(Cbegin);
    rpc_stream.rpc_validate();
    in_transaction = true;
  };
};


void LDB::commit_transaction () {
  if (in_transaction) {
    rpc_stream.send(Ccommit);
    rpc_stream.rpc_validate();
    in_transaction = false;
  };
};


void LDB::abort_transaction () {
  if (in_transaction) {
    rpc_stream.send(Cabort);
    rpc_stream.rpc_validate();
    in_transaction = false;
  };
};


void LDB::process_definition ( const char* definition ) {
  rpc_stream.send(Cprocess);
  rpc_stream.send(String(definition));
  rpc_stream.rpc_validate();
};
