/*********************************************************************************
 *
 *  The Java binding for LDB
 *  Copyright (c) 2003 by Leonidas Fegaras, the University of Texas at
 *     Arlington. All rights reserved.
 *  Programmer: Leonidas Fegaras
 *  Date: 02/07/03
 *
 ********************************************************************************/

import java.io.*;
import java.net.*;


class ServerError extends Throwable {
    public String message;
    public ServerError ( String msg ) { message=msg; }
    public String toString () { return message; }
}


class RPCstream {
    public boolean opened;
    protected Socket socket;
    protected DataOutputStream out;
    protected DataInputStream in;

    protected void error ( String msg ) {
	System.err.println(msg);
	try {
	    out.close();
	    in.close();
	    socket.close();
	    System.exit(1);
	} catch (IOException e) {
	    System.exit(1);
	}
    }

    public RPCstream ( String host, int port ) {
	opened = false;
        try {
	    socket = new Socket(host,port);
            out = new DataOutputStream(socket.getOutputStream());
            in = new DataInputStream(socket.getInputStream());
        } catch (UnknownHostException e) {
            error("Couldn't connect to the host: "+host);
        } catch (IOException e) {
            error("Couldn't get I/O from host: "+host);
        }
    }

    protected void close () {
	if (opened)
	    try {
		out.writeShort(-1);
		out.flush();
		out.close();
		in.close();
		socket.close();
	    } catch (IOException e) {
		error("Cannot close socket");
	    }
    }

    public void send ( String s ) {
	try {
	    out.writeInt(s.length());
	    out.writeBytes(s);
	    out.flush();
	} catch (IOException e) {
	    error("Cannot send string");
	}
    }

    public void send ( int n )  {
	try {
	    out.writeInt(n);
	    out.flush();
	} catch (IOException e) {
	    error("Cannot send string");
	}
    }

    public void send ( short n )  {
	try {
	    out.writeShort(n);
	    out.flush(); 
	} catch (IOException e) {
	    error("Cannot send string");
	}
    }

    public String receiveString () {
	try {
	    int len = in.readInt();
	    byte[] content = new byte[len];
	    in.read(content,0,len);
	    return new String(content);
	} catch (IOException e) {
	    error("Cannot receive string");
	    return "";
	}
    }

    public int receiveInt () {
	try {
	    return in.readInt();
	} catch (IOException e) {
	    error("Cannot receive integer");
	    return 0;
	}
    }

    public int receiveShort () {
	try {
	    return in.readShort();
	} catch (IOException e) {
	    error("Cannot receive short integer");
	    return 0;
	}
    }

    public float receiveFloat () {
	try {
	    return in.readFloat();
	} catch (IOException e) {
	    error("Cannot receive float number");
	    return (float) 0.0;
	}
    }

    public void rpc_validate () throws ServerError {
	try {
	    if (in.readByte() != 0)
	       throw new ServerError(receiveString());
	} catch (IOException e) {
	    error("Cannot validate RPC transmission");
	}
    }
}


class RPCopr {
    protected static short n = 0;
    public final static short Cflush = n++;
    public final static short CgetInt = n++;
    public final static short CgetString = n++;
    public final static short CgetFloat = n++;
    public final static short Csize = n++;
    public final static short Cnth = n++;
    public final static short CtoString = n++;
    public final static short CgetType = n++;
    public final static short Copen = n++;
    public final static short Cclose = n++;
    public final static short Cquery = n++;
    public final static short Cupdate = n++;
    public final static short Cinsert = n++;
    public final static short Cremove = n++;
    public final static short Ccreatei = n++;
    public final static short Cdropi = n++;
    public final static short Cchange = n++;
    public final static short Ctclschema = n++;
    public final static short Cbegin = n++;
    public final static short Ccommit = n++;
    public final static short Cabort = n++;
    public final static short Cprocess = n++;
    public final static short Cstmt = n++;

    public static RPCstream rpc_stream;
}


class LDBobject extends RPCopr {
    protected int ID;

    public LDBobject ( int id ) { ID=id; };

    public LDBobject () {
	ID = rpc_stream.receiveInt();
    }

    public String ref () {
	return "(:" + ID + ")";
    }

    void operation ( short opr ) {
	rpc_stream.send(opr);
	rpc_stream.send(ID);
    }

    public void flush () throws ServerError {
	operation(Cflush);
	rpc_stream.rpc_validate();
    }

    protected void finalize () throws Throwable {
	flush();
	super.finalize();
    }

    public int getInt () throws ServerError {
	operation(CgetInt);
	rpc_stream.rpc_validate();
	return rpc_stream.receiveInt();
    }

    public String getString () throws ServerError {
	operation(CgetString);
	rpc_stream.rpc_validate();
	return rpc_stream.receiveString();
    }

    public float getFloat () throws ServerError {
	operation(CgetFloat);
	rpc_stream.rpc_validate();
	return rpc_stream.receiveFloat();
    }

    public String toString () {
	operation(CtoString);
	try {
	    rpc_stream.rpc_validate();
	    String s = rpc_stream.receiveString();
	    return s;
	} catch (ServerError s) {
	    return s.message;
	}
    }

    public String getType () throws ServerError {
	operation(CgetType);
	rpc_stream.rpc_validate();
	return rpc_stream.receiveString();
    }

    public int size () throws ServerError {
	operation(Csize);
	rpc_stream.rpc_validate();
	return rpc_stream.receiveInt();
    }

    public LDBobject nth ( int index ) throws ServerError {
	operation(Cnth);
	rpc_stream.send(index);
	rpc_stream.rpc_validate();
	return new LDBobject();
    }
}


class LDB extends RPCopr {

    protected boolean opened;
    protected boolean in_transaction;
    final static int port_number = 9737;

    public void open ( String host, String username, String password, String schema ) throws ServerError {
	rpc_stream = new RPCstream(host,port_number);
	rpc_stream.send(Copen);
	rpc_stream.send(username);
	rpc_stream.send(password);
	rpc_stream.send(schema);
	rpc_stream.rpc_validate();
	rpc_stream.opened = true;
	opened = true;
    }

    public void open ( String schema ) throws ServerError {
	open("127.0.0.1","admin","pass",schema);
    }

    public void close () {
	if (opened) {
	    try {
		if (in_transaction)
		    abort_transaction();
		rpc_stream.send(Cclose);
		rpc_stream.rpc_validate();
		rpc_stream.close();
		opened = false;
	    } catch (ServerError m) {
		System.err.println("Cannot terminate client");
		System.exit(1);
	    }
	}
    }

    protected void finalize () throws Throwable {
	close();
	super.finalize();
    }

    public LDBobject query ( String query ) throws ServerError {
	rpc_stream.send(Cquery);
	rpc_stream.send(query);
	rpc_stream.rpc_validate();
	return new LDBobject();
    }

    public void statement ( String stmt ) throws ServerError {
	rpc_stream.send(Cstmt);
	rpc_stream.send(stmt);
	rpc_stream.rpc_validate();
    }

    public void update ( String varname, String query ) throws ServerError {
	rpc_stream.send(Cupdate);
	rpc_stream.send(varname);
	rpc_stream.send(query);
	rpc_stream.rpc_validate();
    }

    public void insert ( String varname, String query ) throws ServerError {
	rpc_stream.send(Cinsert);
	rpc_stream.send(varname);
	rpc_stream.send(query);
	rpc_stream.rpc_validate();
    }

    public void remove ( String varname, String query ) throws ServerError {
	rpc_stream.send(Cremove);
	rpc_stream.send(varname);
	rpc_stream.send(query);
	rpc_stream.rpc_validate();
    }

    public void create_index ( String index_name, String class_name, String attribute ) throws ServerError {
	rpc_stream.send(Ccreatei);
	rpc_stream.send(index_name);
	rpc_stream.send(class_name);
	rpc_stream.send(attribute);
	rpc_stream.rpc_validate();
    }

    public void remove_index ( String index_name ) throws ServerError {
	rpc_stream.send(Cdropi);
	rpc_stream.send(index_name);
	rpc_stream.rpc_validate();
    }

    public void change_password ( String oldp, String newp ) throws ServerError {
	rpc_stream.send(Cchange);
	rpc_stream.send(oldp);
	rpc_stream.send(newp);
	rpc_stream.rpc_validate();
    }

    public String tcl_schema () throws ServerError {
	rpc_stream.send(Ctclschema);
	rpc_stream.rpc_validate();
	return rpc_stream.receiveString();  
    }

    public void begin_transaction () throws ServerError {
	if (!in_transaction) {
	    rpc_stream.send(Cbegin);
	    rpc_stream.rpc_validate();
	    in_transaction = true;
	}
    }

    public void commit_transaction () throws ServerError {
	if (in_transaction) {
	    rpc_stream.send(Ccommit);
	    rpc_stream.rpc_validate();
	    in_transaction = false;
	}
    }

    public void abort_transaction () throws ServerError {
	if (in_transaction) {
	    rpc_stream.send(Cabort);
	    rpc_stream.rpc_validate();
	    in_transaction = false;
	}
    }
}
