#!/bin/sh
# the next line restarts using wish \
exec wish "$0" "$@"
#--------------------------------------------------------------------------------
#
# Tcl/Tk script for ODMG OQL visual query formulation (VOODOO).
# The documentation is available at: http://lambda.uta.edu/ldb/doc/visual.html
# Programmer: Leonidas Fegaras
# Creation: 3/5/99, last update: 2/14/03
# Copyright (c) 1999-2003 by Leonidas Fegaras, the University of Texas at
#     Arlington. All rights reserved.
# This material is provided as is, with absolutely no warranty expressed
# or implied.  Any use is at your own risk.
# Permission is hereby granted to use or copy this program for any
# purpose, provided the above notices are retained on all copies.
#
#--------------------------------------------------------------------------------


source "[file dirname $argv0]/rpc.tcl"


# set this to 1 for debugging
set debug 0


# fonts used in buttons, labels, and text
set vfont -*-times-medium-r-normal--10-*-*-*-*-*-*-*
set bfont -*-times-bold-r-normal--12-*-*-*-*-*-*-*
set tfont -*-courier-medium-r-normal--14-*-*-*-*-*-*-*


# The ODL schema (is read from server).
# Each class has the form:
#   {class_name {optional_superclass} {optional_extent} ...attributes}
set classes {}


# Return the attributes of a class
proc find_class {class_name} {
  global classes
  foreach c $classes {
    if {[lindex $c 0] == $class_name} {
       return $c
    }
  }
  return {}
}


# Return the type of the persistent root (ie. all class extents)
proc persistent_root {} {
  global classes
  set res "struct"
  foreach c $classes {
    if {[llength [lindex $c 2]] == 1} {
       lappend res [list [list "set" [lindex $c 0]] [lindex [lindex $c 2] 0]]
    }
  }
  return $res
}


# Display an error message in a new window
proc voodoo_error {message} {
  #tk_messageBox -message "$message" -title message
  puts "$message"
  exit 1
}


set no_message 0


# Display a help message at the bottom of the screen
proc message {m} {
   global message no_message
   if {$no_message == 0} {
      set message $m
   }
}


proc clear_message {} {
   global message no_message
   if {$no_message == 0} {
      set message " "
   }
}


#--------------------------------------------------------------------------------
# Each object template in the canvas represents one of the following:
#    an ODL object class, a struct, or a simple domain (such as integer).
# Each template is mapped onto an ODL type.
# Each template may have any number of accumulators.
#--------------------------------------------------------------------------------


# Return the environment (variables with their paths) of a template
proc scopevars {o} {
  global object
  set res {}
  foreach c $object($o,scope) {
     lappend res [lindex $c 0]
  }
  return $res
}



# Return the children templates of the template o
proc children {o} {
  global object
  set res {}
  foreach c $object($o,scope) {
     lappend res $o+[lindex $c 0]
  }
  return $res
}


proc remove_join_line {c} {
  global object
  if {[winfo exists .c.join_menu($c)]} {
     .c delete $object($c,join_line)
     destroy .c.join_menu($c)
     .c delete $object($c,join_menu)
  }
}


# Delete the template associated with the window win and all its children and joins
proc delete_object {win} {
  global window_object object
  if {[winfo exists $win]} {
     set o $window_object($win)
     set object($o,head) ""
     set object($o,condition) ""
     destroy $win
     .c delete $object($o,line)
     foreach v [children $o] {
	set object($v,print) 0
	set object($v,group) 0
	set object($v,sort) 0
	foreach d $object($v,joins) {
	   if {[info exists object($d,join)]} {
	      unset object($d,join)
	   }
	   remove_join_line $d
	}
	set object($v,joins) {}
	if {[info exists object($v,join)]} {
	   set r {}
	   foreach x $object($object($v,join),joins) {
	      if {$x != "$v"} {
		 lappend r $x
	      }
	   }
	   set object($object($v,join),joins) $r
	   unset object($v,join)
	}
        remove_join_line $v
	delete_object .c.$v
     }
  }
}


# Draw a join line between the attribute c of template o to (x,y)
proc join_line {o c x y} {
  global object
  set coords [.c coords $object($c,join_line)]
  set fx [expr $object($o,X) + $object($c,linkX)]
  set fy [expr $object($o,Y) + $object($c,linkY)]
  set nx [expr $object($o,X) + $object($c,linkX) + $x - 48]
  set ny [expr $object($o,Y) + $object($c,linkY) + $y - 10]
  .c coords $object($c,join_line) $fx $fy [expr $fx + ($nx-$fx)/2.0] $fy \
		[expr $fx + ($nx-$fx)/2.0] $ny $nx $ny
  .c coords $object($c,join_menu) [expr $fx + ($nx-$fx)/2.0] [expr $fy + ($ny-$fy)/2.0]
}


proc join_menu {c op} {
  global object
  set object($c,join_cond) $op
}


proc find_button_path {w} {
   set a [split $w .]
   if {[llength $a] == 4} {
      set o [lindex [split [lindex $a 2] {()}] 1]
      set v [lindex [split [lindex $a 3] {(),}] 1]
      if {.c.object($o).attribute($v,button) == $w} {
	 return $o+$v
      } else {
	 return ""
      }
   } else {
      return ""
   }
}


proc bind_join_motion {W X Y x y} {
  global window_object object last_join
  clear_label
  set win [winfo parent $W]
  set c $window_object($win)
  set o $object($c,parent)
  if {$o != "root"} {
     if {![winfo exists .c.join_menu($c)]} {
	set object($c,join_line) [.c create line 0 0 0 0 0 0 0 0 -fill red -width 2]
	set mc .c.join_menu($c).choices
	menubutton .c.join_menu($c) -text "=" -textvariable object($c,join_cond) \
   		 -width 2 -menu $mc -bg turquoise -activebackground turquoise -bd 1 -relief raised
	menu $mc -bg PaleTurquoise1 -activebackground turquoise -tearoff 0
	$mc add command -label "=" -command "join_menu $c ="
	$mc add command -label "!=" -command "join_menu $c !="
	$mc add command -label ">" -command "join_menu $c >"
	$mc add command -label "<" -command "join_menu $c <"
	$mc add command -label ">=" -command "join_menu $c >="
	$mc add command -label "<=" -command "join_menu $c <="
	set object($c,join_menu) [.c create window 0 0 -window .c.join_menu($c)]
        bind .c.join_menu($c) <Any-Enter> {
	   message "select a join comparison operator"
	}
        bind .c.join_menu($c) <Any-Leave> {
	   clear_message
	}
     }
     set w [winfo containing $X $Y]
     if {[regexp {\.c\..*\..*\.button} $w] \
		&& [winfo parent $w] != "$win" \
		&& [winfo parent [winfo parent $w]] != ".c.root" \
		&& [type_of $c] == [type_of $window_object([winfo parent $w])] \
		&& [valid_path $c $window_object([winfo parent $w])] > 0} {
        set win [winfo parent $w]
        set e $window_object($win)
        set d $object($e,parent)
        set nx [expr $object($o,X) + $object($c,linkX) - 48]
        set ny [expr $object($o,Y) + $object($c,linkY) - 10]
        join_line $o $c [expr $object($d,X) + $object($e,linkX) - $nx] \
  		        [expr $object($d,Y) + $object($e,linkY) - $ny]
        .c itemconfigure $object($c,join_line) -fill green
	set last_join [list $o $c $d $e]
     } else {
        join_line $o $c $x $y
        .c itemconfigure $object($c,join_line) -fill red
        set last_join [list $o $c]
     }
  }
}


proc ancestors {o} {
  return [split $o +]
}


proc no_accumulators_in_path {r s} {
  global object
  set path [join $r +]
  foreach a $s {
     set path $path+$a
     if {$object($path,accumulators) != {}} {
	return 0
     }
  }
  return 1
}


proc valid_path {c d} {
  set p1 [ancestors $c]
  set p2 [ancestors $d]
  set i 0
  foreach a $p1 b $p2 {
     if {$a != $b} {
	break
     }
     incr i
  }
  set r [join [lrange $p1 0 [expr $i-1]] +]
  if {[no_accumulators_in_path $r [lrange $p1 $i [expr [llength $p1]-2]]]} {
     return 1
  } elseif {[no_accumulators_in_path $r [lrange $p2 $i [expr [llength $p2]-2]]]} {
     return 2
  } else {
     return 0
  }
}


proc bind_join_release {} {
   global last_join object
   if {[llength $last_join] == 2} {
      remove_join_line [lindex $last_join 1]
      if {[info exists object([lindex $last_join 1],join)]} {
	 set o [lindex $last_join 1]
	 set r {}
	 foreach x $object($o,joins) {
	    if {$x != $o} {
	       lappend r $x
	    }
	 }
	 set object($o,joins) $r
	 unset object($o,join)
      }
   } elseif {[llength $last_join] == 4} {
      set object([lindex $last_join 1],join) [lindex $last_join 3]
      lappend object([lindex $last_join 3],joins) [lindex $last_join 1]
      set last_join {}
   }
}


proc find_object {win} {
  global window_object
  if {[info exists window_object($win)]} {
     if {[info exists window_object([winfo parent $win])]} {
	return $window_object([winfo parent $win])
     }
     return $window_object($win)
  } else {
     return [find_object [winfo parent $win]]
  }
}


proc bind_button_motion {W X Y x y} {
  global window_object object last_copy source_object destination_object vfont
  clear_label
  set win [winfo parent $W]
  set c $window_object($win)
  if {![winfo exists .c.copy]} {
     label .c.copy -text $object($c,name) -fg white -bg red -font $vfont
     set last_copy [.c create window 0 0 -window .c.copy]
  }
  set sx [.c canvasx [expr [winfo pointerx $W]-[winfo rootx .c]]]
  set sy [.c canvasy [expr [winfo pointery $W]-[winfo rooty .c]]]
  .c coords $last_copy $sx [expr $sy+20]
  raise .c.copy
  set w [winfo containing $X $Y]
  if {[string last ".c.$object($c,parent)" "$w"] == 0} {
     .c.copy config -bg green
     set source_object $c
     set destination_object [find_object $w]
  } else {
     .c.copy config -bg red
     set source_object "none"
  }
}


proc find_path {v s} {
  foreach a $s {
     if {[lindex $a 0] == $v} {
	return [lindex $a 1]
     }
  }
}


proc bind_button_release {W} {
  global object window_object last_copy source_object destination_object
  if {[winfo exists .c.copy]} {
     destroy .c.copy
     .c delete $last_copy
  }
  if {![info exists source_object] || $source_object == ""} {
     set source_object $window_object([winfo parent $W])
     set destination_object $object($source_object,parent)
  }
  if {[info exists source_object] && $source_object != "none"} {
     set name [string trimright $object($source_object,name) "'"]
     while {[find $name $object($destination_object,scope)]} {
	set name "${name}'"
     }
     set p [find_path $object($source_object,name) $object($object($source_object,parent),scope)]
     if {[info exists object($p,copies)]} {
	incr object($p,copies)
     } else {
	set object($p,copies) 1
     }
     display_attribute $destination_object $name $object($source_object,otype) $p $object($p,copies) 0 0
  }
  set source_object ""
}


proc print_type {type} {
  if {[lindex $type 0] == "struct"} {
     if {[llength $type] == 2} {
        return "[print_type [lindex [lindex $type 1] 0]]"
     }
     set res "struct( "
     foreach x [lrange $type 1 end] {
	set res "${res}[lindex $x 1]: [print_type [lindex $x 0]], "
     }
     if {[string length $res] == 8} {
	return "struct()"
     } else {
	return "[string range $res 0 [expr [string length $res]-3]] )"
     }
  } elseif {[lindex $type 0] == "enum"} {
     return "enum([lindex $type 1])"
  } elseif {[lindex $type 0] == "union"} {
     return "union([lindex $type 1])"
  } elseif {[llength $type] == 2} {
     return "[lindex $type 0]([print_type [lindex $type 1]])"
  }
  return "$type"
}


proc attribute_name {o} {
  global object
  if {$o == "root"} {
     return $o
  }
  set res $object($o,name)
  set x $object($o,parent)
  while {$x != "root" && $object($x,accumulators) == {}} {
    set res "$object($x,name)\_$res"
    set x $object($x,parent)
  }
  return $res
}


proc type_of_path {o} {
  global object
  if {[info exists object($o,accumulators)]} {
     return [type_of_path_acc $o $object($o,accumulators)]
  } else {
     return [type_of_path_acc $o {}]
  }
}


proc type_of_path_acc {o accumulators} {
  global object
  if {[winfo exists .c.$o]} {
     set unionp 0
     if {[lindex $object($o,otype) 0] == "union"} {
        set unionp 1
     }
     set res {struct}
     foreach c [scopevars $o] {
	if {$unionp || $object($o+$c,print)} {
	   set tp [type_of_path $o+$c]
	   if {[lindex $tp 0] == "struct" && [llength $tp] == 2} {
	      lappend res [lindex $tp 1]
	   } else {
	      lappend res [list $tp [attribute_name $o+$c]]
	   }
	}
     }
     if {$unionp} {
        set utp [lindex [lindex $res 1] 0]
        foreach a [lrange $res 2 end] {
	   if {$utp != [lindex $a 0]} {
	      return error
	   }
        }
	set res $utp
     }
     if {!$unionp && [info exists object($o,head)] && $object($o,head) != ""} {
	lappend res {? head}
     }
     if {!$unionp && [llength $res] == 1} {
	if {[info exists object($o,condition)] && $object($o,condition) != ""} {
	   set res bool
	} else {
	   set tp $object($o,gtype)
	   foreach v $object($o,accumulators) {
		set tp [lindex $tp 1]
	   }
	   lappend res [list $tp [attribute_name $o]]
	}
     }
     set rs $res
     foreach v $accumulators {
        set rs [remove_trivial_struct $res]
	if {$v == "all" || $v == "some"} {
	   if {$rs == "?"} {
	      set rs "bool"
	   }
	   if {$rs != "bool"} {
	      return "none"
	   }
       } elseif {$v == "count"} {
	   set rs "int"
       } elseif {$v == "sum" || $v == "avg" || $v == "max" || $v == "min"} {
	   if {$rs == "?"} {
	      set rs "int"
	   }
	   if {$rs != "int" && $rs != "long" && ($rs != "string" || $v == "sum" || $v == "avg")} {
	      return "none"
	   }
	} elseif {$v != "each" && $v != "elem"} {
	   set rs [list $v $rs]
	}
     }
     return $rs
  } else {
     return $object($o,gtype)
  }
}


proc remove_trivial_struct {type} {
  if {[lindex $type 0] == "struct"} {
     if {[llength $type] == 2} {
        return "[lindex [lindex $type 1] 0]"
     }
     return $type
  } elseif {[llength $type] == 2} {
     if {[lindex [lindex $type 1] 0] == "struct"} {
        if {[llength [lindex $type 1]] == 2} {
           return [list [lindex $type 0] [lindex [lindex [lindex $type 1] 1] 0]]
        }
     }
     return $type
  }
  return $type
}


proc type_of {o} {
  return [subst [remove_trivial_struct [type_of_path $o]] "'" "_"]
}


proc display_label {text W} {
  global type_label vfont no_message
  if {$no_message == 0} {
     label .c.label -text $text -bg green -bd 3 -relief raised -font $vfont -justify left -wraplength 300
     set sx [.c canvasx [expr [winfo pointerx $W]-[winfo rootx .c]]]
     set sy [.c canvasy [expr [winfo pointery $W]-[winfo rooty .c]]]
     set type_label [.c create window [expr $sx+10] [expr $sy+10] -window .c.label -anchor nw]
  }
}


proc bind_type_label {W} {
  global window_object object
  set win [winfo parent $W]
  set c $window_object($win)
  display_label [print_type [type_of $c]] $W
}


proc clear_label {} {
  global type_label
  .c delete $type_label
  if {[winfo exists .c.label]} {
     destroy .c.label
  }
}


proc find {v s} {
  foreach x $s {
     if {[lindex $x 0] == $v} {
	return 1
     }
  }
  return 0
}


proc remove {v s} {
  set res {}
  foreach x $s {
     if {[lindex $x 0] != $v} {
	lappend res $x
     }
  }
  return $res
}


proc handle_partitions {w} {
  global object window_object
  set c $window_object([winfo parent $w])
  set v $object($c,name)
  set o $object($c,parent)
  set oper [lindex [split $w .] 4]
  set already_grouped [find "partition" $object($o,scope)]
  if {$object($c,$oper)} {
     .c.$o.$c.$oper config -bg LightSkyBlue -fg blue
     set object($c,$oper) 0
  } else {
     .c.$o.$c.$oper config -bg blue -fg LightBlue
     set object($c,$oper) 1
  }
  set group_by 0
  foreach v [children $o] {
     if {$object($v,group)} {
	set group_by 1
     }
  }
  set tp [type_of $o]
  if {$tp == "none" || (($tp == "int" || $tp == "long" || $tp == "bool" || $tp == "?") \
			&& $already_grouped != $group_by)} {
     message "can't push this button (invalid operation)"
     if {$object($c,$oper)} {
	.c.$o.$c.$oper config -bg LightSkyBlue -fg blue
	set object($c,$oper) 0
     } else {
	.c.$o.$c.$oper config -bg blue -fg LightBlue
	set object($c,$oper) 1
     }
     return
  }
  if {$already_grouped} {
     if {!$group_by} {
	foreach a [scopevars $o] {
	   delete_object .c.$o+$a
	   if {!$object($o+$a,group) && $a != "partition" && $a != $v} {
	      set object($o+$a,gtype) $object($o+$a,otype)
	   }
	   if {[regexp {partition'*} $a]} {
	      delete_object .c.$o+$a
	      destroy .c.$o.$o+$a
	      set object($o,scope) [remove $a $object($o,scope)]
	   }
	}
     } elseif {$oper == "group"} {
	if {$object($c,group)} {
	   set object($c,gtype) $object($c,otype)
	} else {
	   set object($c,gtype) [list "set" $object($c,otype)]
	}
     }
  } else {
     if {$group_by} {
	foreach a [scopevars $o] {
	   delete_object .c.$o+$a
	   if {!$object($o+$a,group) && $a != "partition" && $a != $v} {
	      set object($o+$a,gtype) [list "set" $object($o+$a,otype)]
	   }
	}
	display_attribute $o "partition" $object($o,otype) $o 0 0 0
     } elseif {$oper == "group"} {
	if {$object($c,group)} {
	   set object($c,gtype) $object($c,otype)
	} else {
	   set object($c,gtype) [list "set" $object($c,otype)]
	}
     }
  }
}


proc create_template {w} {
  global object window_object
  set c $window_object([winfo parent $w])
  set v $object($c,name)
  set o $object($c,parent)
  set s [find_path $v $object($o,scope)]
  display_frame $o $v $s+$v
}


proc display_attribute {o v type source_template copy_num group unionp} {
   global object window_object vfont
   set w .c.$o.$o+$v
   lappend object($o,scope) [list $v $source_template $copy_num]
   set object($o+$v,otype) $type
   set object($o+$v,gtype) $type
   set object($o+$v,cond) 0
   set object($o+$v,print) $unionp
   set object($o+$v,group) 0
   set object($o+$v,sort) 0
   set object($o+$v,name) $v
   set object($o+$v,scope) {}
   set object($o+$v,joins) {}
   set window_object(.c.$o.$o+$v) $o+$v
   frame $w -bg turquoise
   pack $w -in .c.$o -pady 1
   set name $v
   if {[string length $v] > 18} {
      set name [string range $v 0 17]
   }
   label $w.button -text $name -width 16 -bg LightSkyBlue -bd 2 -relief raised -font $vfont
   if {$unionp == 0} {
      label $w.print -text "P" -width 1 -padx 3 -bg LightSkyBlue -fg blue -bd 1 -relief raised -font $vfont
   } else {
      label $w.print -text " " -width 1 -padx 3 -bg turquoise -bd 1 -font $vfont
   }
   if {[llength $object($o,accumulators)] > 0} {
      if {$group == 0} {
	 $w.button configure -fg blue
	 label $w.group -text " " -width 1 -padx 3 -bg turquoise -bd 1 -font $vfont
	 if {$object($o,sorted)} {
	    label $w.sort -text " " -width 1 -padx 3 -bg turquoise -bd 1 -font $vfont
	    pack $w.print $w.group $w.sort $w.button -in $w -side left -padx 1
	 } else {
	    pack $w.print $w.group $w.button -in $w -side left -padx 1
	 } 
      } else {
	 label $w.group -text "G" -width 1 -padx 3 -bg LightSkyBlue -fg blue -bd 1 -relief raised -font $vfont
	 pack $w.print $w.group $w.button -in $w -side left -padx 1
	 bind $w.group <Button-1> {
	    handle_partitions %W
	 }
	 bind $w.group <Enter> {
	    message "select this attribute to group by"
	 }
	 bind $w.group <Leave> {
	    clear_message
	 }
      }
   } else {
      if {$group == 0} {
	 $w.button configure -fg blue
      }
      pack $w.print $w.button -in $w -side left -padx 1
   }
   update
   update idletasks
   set object($o+$v,parent) $o
   set object($o+$v,linkX) [expr [winfo width $w] / 2.0 + 3]
   set object($o+$v,linkY) [expr [winfo y $w] + [winfo height $w] / 2.0]
   if {!$unionp} {
      bind $w.print <Button-1> {
         handle_partitions %W
      }
      bind $w.button <B1-Motion> {
         global joinp
         set joinp 1
         bind_join_motion %W %X %Y %x %y
         message "to make a join, leave the mouse on top of an attribute that makes the join line green"
      }
      bind $w.button <B1-ButtonRelease> {
         global joinp
         if {[info exist joinp] && $joinp} {
   	 bind_join_release
         } else {
   	 create_template %W
         }
         set joinp 0
         clear_message
      }
      bind $w.button <B2-ButtonRelease> {
         bind_button_release %W
         clear_message
      }
      bind $w.print <Enter> {
        message "select this attribute to appear in the template value"
      }
      bind $w.print <Leave> {
        clear_message
      }
   } else {
      bind $w.button <B1-ButtonRelease> {
   	 create_template %W
         clear_message
      }
   }
   if {$unionp} {
      bind $w.button <Enter> {
        bind_type_label %W
        message "left: create the template (note: union cases must be of the same type)"
      }
   } elseif {![regexp {partition'*} $v]} {
      bind $w.button <B2-Motion> {
         bind_button_motion %W %X %Y %x %y
         message "to copy this attribute, leave it on top of a template"
      }
      bind $w.button <Enter> {
        bind_type_label %W
        message "left: create the template,   middle: copy the attribute,   left-drag: make a join,   middle-drag: copy the attribute in a template"
      }
   } else {
      bind $w.button <Button-2> {
         bind_button_motion %W %X %Y %x %y
      }
      bind $w.button <Enter> {
        bind_type_label %W
        message "left: create the template,   middle: copy the attribute,   left-drag: make a join"
      }
   }
   bind $w.button <Leave> {
     clear_label
     clear_message
   }
}


# Display the attributes of object o inside the o template
proc display_attributes {o tp classp} {
  global window_object object
  if {$classp} {
     set class [find_class $tp]
     set super [lindex $class 1]
     set attributes [lrange $class 3 end]
     if {$super != {}} {
        display_attributes $o [lindex $super 0] 1
     }
  } else {
     set class $tp
     set attributes [lrange $class 1 end]
  }
  foreach a $attributes {
     display_attribute $o [lindex $a 1] [lindex $a 0] $o 0 1 0
  }
}


proc make_header {o condp} {
  global object vfont
  set w .c.$o
  frame $w.headf -bg turquoise
  pack $w.headf -expand 1
  if {$condp} {
     label $w.headf.headl -text "head" -bg turquoise -fg blue -font $vfont
     entry $w.headf.hvalue -width 15 -bg white -textvariable object($o,head) -font $vfont
     pack $w.headf.headl $w.headf.hvalue -side left -fill x -pady 2 -expand 1
     frame $w.condf -bg turquoise
     pack $w.condf -expand 1
     bind $w.headf <Enter> {
        message "write an OQL expression here (the mapped value is \$)"
     }
     bind $w.headf <Leave> {
        clear_message
     }
     label $w.condf.condl -text "cond" -bg turquoise -fg blue -font $vfont
     entry $w.condf.cond -width 15 -bg white -textvariable object($o,condition) -font $vfont
     pack $w.condf.condl $w.condf.cond -side left -fill x -pady 2 -expand 1
     bind $w.condf <Enter> {
	message "write a condition here (eg. >100)"
     }
     bind $w.condf <Leave> {
	clear_message
     }
  }
}


proc display_class {o class_name setp} {
  global object
  if {$setp} {
     make_header $o 0
  }
  frame .c.$o.attrs -bg turquoise
  pack .c.$o.attrs -in .c.$o
  display_attributes $o $class_name 1
}


proc display_struct {o tp} {
  frame .c.$o.attrs -bg turquoise
  pack .c.$o.attrs -in .c.$o
  display_attributes $o $tp 0
}


proc display_union {o tp} {
  frame .c.$o.attrs -bg turquoise
  pack .c.$o.attrs -in .c.$o
  set attributes [lrange $tp 3 end]
  foreach a $attributes {
     display_attribute $o [lindex $a 2] [lindex $a 1] $o 0 1 1
  }
}


proc bind_template_motion {o x y subtree} {
   global object
   clear_label
   set ww $object($o,window)
   set object($o,X) [expr $object($o,X)+$x]
   set object($o,Y) [expr $object($o,Y)+$y]
   .c coords $ww $object($o,X) $object($o,Y)
   if {$o != "root"} {
      set coords [.c coords $object($o,line)]
      .c coords $object($o,line) $object($o,X) [expr $object($o,Y) + 20] \
			         [lindex $coords 2] [lindex $coords 3]
   }
   foreach c [children $o] {
      if {[info exists object($c,join)]} {
	 set d $object($c,join)
	 set fx [expr $object($o,X) + $object($c,linkX)]
	 set fy [expr $object($o,Y) + $object($c,linkY)]
	 set nx [expr $object($object($d,parent),X) + $object($d,linkX)]
	 set ny [expr $object($object($d,parent),Y) + $object($d,linkY)]
	 .c coords $object($c,join_line) $fx $fy [expr $fx + ($nx-$fx)/2.0] $fy \
					 [expr $fx + ($nx-$fx)/2.0] $ny $nx $ny
	.c coords $object($c,join_menu) [expr $fx + ($nx-$fx)/2.0] [expr $fy + ($ny-$fy)/2.0]
      }
     foreach d $object($c,joins) {
	set nx [expr $object($object($d,parent),X) + $object($d,linkX)]
	set ny [expr $object($object($d,parent),Y) + $object($d,linkY)]
	set fx [expr $object($o,X) + $object($c,linkX)]
	set fy [expr $object($o,Y) + $object($c,linkY)]
	.c coords $object($d,join_line) $fx $fy [expr $fx + ($nx-$fx)/2.0] $fy \
					[expr $fx + ($nx-$fx)/2.0] $ny $nx $ny
	.c coords $object($d,join_menu) [expr $fx + ($nx-$fx)/2.0] [expr $fy + ($ny-$fy)/2.0]
      }
      if {[winfo exists .c.$c]} {
	 set coords [.c coords $object($c,line)]
	 .c coords $object($c,line) [lindex $coords 0] [lindex $coords 1] \
				    [expr $object($o,X) + $object($c,linkX) * 2.0] \
				    [expr $object($o,Y) + $object($c,linkY)]
	 if {$subtree} {
	    bind_template_motion $c $x $y 1
	 }
      }
   }
}


proc handle_order_by {w} {
  global object window_object
  set o $window_object([winfo parent $w])
  if {$object($o,sort)} {
     $w config -bg LightSkyBlue -fg blue
     set object($o,sort) 0
  } else {
     $w config -bg blue -fg LightSkyBlue
     set object($o,sort) 1
  }
}


proc acc_menu {acc o i sortp} {
  global object vfont
  set old_accs $object($o,accumulators)
  set object($o,accumulators) [lreplace $object($o,accumulators) $i $i $acc]
  if {[type_of $o] == "none"} {
     set object($o,accumulators) $old_accs
     return
  }
  .c.$o.head.menu($i) configure -text $acc
  if {$sortp && $i == 0 && !$object($o,sorted)} {
     set object($o,sorted) 1
     foreach v [children $o] {
	set object($v,linkX) [expr $object($v,linkX) + 9]
	if {$object($v,name) == "partition"} {
	   label .c.$o.$v.sort -text " " -width 1 -padx 3 -bg turquoise -bd 0 -font $vfont
	   pack .c.$o.$v.sort -before .c.$o.$v.button -side left -padx 2
        } else {
	   set object($v,sort) 0
	   label .c.$o.$v.sort -text "S" -width 1 -padx 3 -bg LightSkyBlue \
	       -fg blue -bd 1 -relief raised -font $vfont
	   pack .c.$o.$v.sort -before .c.$o.$v.button -side left -padx 1
	   bind .c.$o.$v.sort <Button-1> {
	      handle_order_by %W
	   }
	   bind .c.$o.$v.sort <Enter> {
	      message "select this attribute to sort"
	   }
	   bind .c.$o.$v.sort <Leave> {
	      clear_message
	   }
	}
    }
  } elseif {!$sortp && $i == 0 && $object($o,sorted)} {
     set object($o,sorted) 0
     foreach v [children $o] {
        set object($v,sort) 0
	set object($v,linkX) [expr $object($v,linkX) - 9]
	if {[winfo exists .c.$o.$v.sort]} {
	   destroy .c.$o.$v.sort
	}
     }
  }
}


proc deactivate_menu_items {menu o accumulators} {
  set tp [remove_trivial_struct [type_of_path_acc $o $accumulators]]
  if {$tp == "bool" || $tp == "?"} {
     $menu entryconfigure 9 -state normal
     $menu entryconfigure 10 -state normal
  } else {
     $menu entryconfigure 9 -state disabled
     $menu entryconfigure 10 -state disabled
  }
  if {$tp == "int" || $tp == "long" || $tp == "?"} {
     $menu entryconfigure 5 -state normal
     $menu entryconfigure 6 -state normal
     $menu entryconfigure 7 -state normal
     $menu entryconfigure 8 -state normal
  } else {
     $menu entryconfigure 5 -state disabled
     $menu entryconfigure 6 -state disabled
     $menu entryconfigure 7 -state disabled
     $menu entryconfigure 8 -state disabled
  }
}


proc display_frame {parent v path} {
  global object window_object vfont bfont
  set o $parent+$v
  set w .c.$o
  if {![winfo exists $w]} {
     set window_object($w) $o
     set object($o,X) [expr $object($parent,X) + $object($o,linkX) * 2.0 + 30]
     set object($o,Y) [expr $object($parent,Y) + $object($o,linkY) - 20]
     set object($o,line) [.c create line $object($o,X) [expr $object($o,Y) + 20] \
				[expr $object($parent,X) + $object($o,linkX) * 2.0] \
				[expr $object($parent,Y) + $object($o,linkY)] \
				-fill blue -width 2.0]
     set object($o,parent) $parent
     set object($o,scope) {}
     set object($o,sorted) 0
     set object($o,accumulators) {}
     frame $w -bg turquoise -bd 4 -relief raised
     set object($o,window) [.c create window $object($o,X) $object($o,Y) -anchor nw -window $w]
     frame $w.head -bg turquoise
     pack $w.head -in $w
     set can_be_elem [expr [llength $object($parent,accumulators)] > 0]
     if {$object($o,name) == "partition" || [find "partition" $object($parent,scope)]} {
        set can_be_elem 0
     }
     set tp $object($o,gtype)
     if {[llength $tp] == 1 && [find_class $tp] == {}} {
	set object($o,cond) 1
     } else {
	set object($o,print) 1
	if {[type_of $parent] == "none" || [.c.$parent.$o.print cget -text] == " "} {
	   set object($o,print) 0
	} else {
	   .c.$parent.$o.print config -bg blue -fg LightBlue
	}
     }
     set i 0
     set setp 0
     while {([llength $tp] == 2) && ([lindex $tp 0] != "struct")} {
	set setp 1
	if {$can_be_elem || $i>0} {
	   set acc "each"
	} else {
	   set acc "bag"
	}
        set tp [lindex $tp 1]
	set mc $w.head.menu($i).choices
        menubutton $w.head.menu($i) -text $acc -width 3 -font $vfont -menu $mc \
   		   -bg turquoise -activebackground turquoise -bd 1 -relief raised
        menu $mc -postcommand "deactivate_menu_items $mc $o [list $object($o,accumulators)]" \
	         -bg PaleTurquoise1 -activebackground turquoise -tearoff 0
	lappend object($o,accumulators) $acc
	if {$can_be_elem || $i>0} {
	   set each_state normal
	} else {
	   set each_state disabled
	}
	$mc add command -label "each" -font $vfont -state $each_state -command "acc_menu each $o $i 0"
        $mc add command -label "set" -font $vfont -command "acc_menu set $o $i 0"
        $mc add command -label "bag" -font $vfont -command "acc_menu bag $o $i 0"
        $mc add command -label "ordered list" -font $vfont -command "acc_menu list $o $i [expr $i==0]"
        $mc add command -label "count" -font $vfont -command "acc_menu count $o $i 0"
        $mc add command -label "summation" -font $vfont -command "acc_menu sum $o $i 0"
        $mc add command -label "average" -font $vfont -command "acc_menu avg $o $i 0"
        $mc add command -label "maximum" -font $vfont -command "acc_menu max $o $i 0"
        $mc add command -label "minimum" -font $vfont -command "acc_menu min $o $i 0"
        $mc add command -label "some" -font $vfont -command "acc_menu some $o $i 0"
        $mc add command -label "all" -font $vfont -command "acc_menu all $o $i 0"
	$mc add command -label "element" -font $vfont -command "acc_menu elem $o $i 0"
        pack $w.head.menu($i) -in $w.head -side left -padx 1m -fill x
        bind $w.head.menu($i) <Any-Enter> {
	   message "select an accumulator"
	}
        bind $w.head.menu($i) <Any-Leave> {
	   clear_message
	}
        incr i
     }
     if {[lindex $tp 0] == "union"} {
        label $w.head.name -text "union" -bg turquoise -font $vfont
        pack $w.head.name -in $w.head -side left -padx 1m -fill x
        display_union $o $tp
     } elseif {[lindex $tp 0] == "struct"} {
        label $w.head.name -text "struct" -bg turquoise -font $vfont
        pack $w.head.name -in $w.head -side left -padx 1m -fill x
        display_struct $o $tp
     } elseif {[find_class $tp] != {}} {
        label $w.head.name -text "$tp" -bg turquoise -font $bfont
        pack $w.head.name -in $w.head -side left -padx 1m -fill x
        display_class $o $tp $setp
     } else {
        label $w.head.name -text "$tp" -bg turquoise -font $vfont
        pack $w.head.name -in $w.head -side left -padx 1m -fill x
	make_header $o 1
     }
     bind $w.head.name <B1-Motion> {
       global window_object no_message
       set no_message 1
       set win [winfo parent [winfo parent %W]]
       raise $win
       set o $window_object($win)
       bind_template_motion $o %x %y 0
     }
     bind $w.head.name <B2-Motion> {
       global window_object no_message
       set no_message 1
       set win [winfo parent [winfo parent %W]]
       raise $win
       set o $window_object($win)
       bind_template_motion $o %x %y 1
     }
     bind $w.head.name <Button-1> {
       raise [winfo parent [winfo parent %W]]
     }
     bind $w.head.name <B1-ButtonRelease> {
       global no_message
       set no_message 0
     }
     bind $w.head.name <B2-ButtonRelease> {
       global no_message
       set no_message 0
     }
     bind $w.head.name <Button-3> {
       global object windo_object
       if {[winfo exists .c.label]} {
	  destroy .c.label
       }
       set w [winfo parent [winfo parent %W]]
       set o $window_object($w)
       set p $object($o,parent)
       set object($o,print) 0
       if {[.c.$p.$o.print cget -text] != " "} {
          .c.$p.$o.print config -bg LightSkyBlue -fg blue
       }
       delete_object $w
       clear_message
     }
     bind $w.head.name <Any-Enter> {
       bind_type_label [winfo parent %W]
       message "left-drag: move template,   middle-drag: move template and its children,   right: remove template"
     }
     bind $w.head.name <Any-Leave> {
       clear_label
       clear_message
     }
     bind $w <Button-1> {
       raise %W
     }
     bind $w <Any-Enter> {
       %W.head.name configure -bg PaleTurquoise1
     }
     bind $w <Any-Leave> {
       %W.head.name configure -bg turquoise
     }
  }
}


proc Join {r m s} {
  if {$r == {}} {
     return {}
  } elseif {[llength $r] == 1} {
     return "[lindex [lindex $r 0] 0]${m}[lindex [lindex $r 0] 1]"
  } else {
     return "[lindex [lindex $r 0] 0]${m}[lindex [lindex $r 0] 1]${s}[Join [lrange $r 1 end] $m $s]"
  }
}


proc make_oql {c} {
  global debug
  if {$debug} {
     puts "&&& $c"
  }
  set acc [lindex $c 0]
  set heads [lindex $c 1]
  set iterations [lindex $c 2]
  set predicates [lindex $c 3]
  set order [lindex $c 4]
  set groupby [lindex $c 5]
  set having [lindex $c 6]
  if {[llength $heads] == 1} {
     set head [lindex [lindex $heads 0] 1]
  } else {
     set head "struct( [Join $heads {: } { , }] )"
  }
  if {$iterations == {}} {
     if {$debug} {
	puts "--> $head"
     }
     return $head
  }
  set predicate "[join $predicates " and "] "
  set iteration [Join $iterations " in " ", "]
  set q "$head\nfrom $iteration"
  set simpl 1
  if {$predicates != {}} {
     set simpl 0
     set q "$q\nwhere $predicate"
  }
  if {$acc == "set"} {
     set q "select distinct $q"
  } else {
     set q "select $q"
  }
  if {$predicates == {} && $groupby == {} && $order == {} && [llength $iterations] == 1 \
		&& $head == [lindex [lindex $iterations 0] 0]} {
     set q [lindex [lindex $iterations 0] 1]
  }
  if {$groupby != {}} {
     set simpl 0
     set q "$q group by [Join $groupby {: } ", "]"
     if {$having != {}} {
	set q "$q having [join $having " and "]"
     }
  }
  if {$order != {}} {
     set simpl 0
     set q "$q order by [join $order ", "]"
  }
  if {$acc == "some" && [llength $iterations] == 1 && $simpl} {
     set q "exists [lindex [lindex $iterations 0] 0] in [lindex [lindex $iterations 0] 1]: $head "
  } elseif {$acc == "all" && [llength $iterations] == 1 && $simpl} {
     set q "for all [lindex [lindex $iterations 0] 0] in [lindex [lindex $iterations 0] 1]: $head "
  } elseif {$acc == "some"} {
     set q "exists X in ($q): X"
  } elseif {$acc == "all"} {
     set q "for all X in ($q): X" 
  } elseif {$acc == "sum"} {
     set q "sum($q)"
  } elseif {$acc == "avg"} {
     set q "avg($q)"
  } elseif {$acc == "count"} {
     set q "count($q)"
  } elseif {$acc == "max"} {
     set q "max($q)"
  } elseif {$acc == "min"} {
     set q "min($q)"
  } elseif {$acc == "elem"} {
     set q "element($q)"
  }
  set q "($q)"
  if {$debug} {
     puts "&-> [subst [join $q] " root." " "]"
  }
  return [subst [subst [subst [join $q] " root." " "] "'." "."] "':" "_:"]
}


proc compile_path {path env has_groupby} {
  global debug
  set pl [split $path "+"]
  set cpath ""
  set pos 0
  foreach v $env p $pl {
     if {$p != [lindex $v 0]} {
	break
     }
     if {[lindex $v 1] == {}} {
	set cpath "${cpath}$p."
     } elseif {[regexp {partition'*} [lindex $v 0]]} {
	 if {[lindex [lindex $v 1] 0] != "PARTITION"} {
	    set cpath "[lindex [lindex $v 1] 0].$cpath"
	 }
     } else {
	set cpath "[lindex [lindex $v 1] 0]."
     }
     incr pos
  }
  if {$pos == [llength $pl]} {
     set cpath [string trimright $cpath '.']
  } elseif {$pos < [llength $pl]} {
     if {[lindex $pl $pos] == "partition"} {
        set cpath "partition"
     } else {
        set cpath "${cpath}[join [lrange $pl $pos end] "."]"
     }
  } else {
     set cpath [join $pl "."]
  }
  if {$debug} {
     puts "* $path   $env    [string trim $cpath "'"]"
  }
  return [string trim $cpath "'"]
}


proc new_name {} {
  global counter
  if {[info exists counter]} {
     incr counter
     return "X_$counter"
  } else {
     set counter 1
     return "X_0"
  }
}


proc subst {s from to} {
  set n [string first $from $s]
  if {$n >= 0} {
     set ns "[string range $s 0 [expr $n-1]]${to}[string range $s [expr $n+[string length $from]] end]"
     return [subst $ns $from $to]
  }
  return $s
}


proc split_ands {pred} {
  set res {}
  set n 0
  while {$n != -1} {
    set n [string first " and " $pred]
    if {$n != -1} {
       lappend res [string range $pred 0 [expr $n-1]]
       set pred [string range $pred [expr $n+5] end]
    } else {
       lappend res $pred
    }
  }
  return $res
}


proc make_predicate {pred path} {
  set res ""
  foreach p [split_ands $pred] {
    if {[regexp {^[a-zA-Z].*} $p]} {
       set res "$res and $path = \'$p\'"
    } elseif {[regexp {^[0-9].*} $p]} {
       set res "$res and $path = $p"
    } elseif {[regexp {^(=|>|<|!).*} $p]} {
       set res "$res and $path $p"
    } else {
       set res "$res and $p"
    }
  }
  return [string range $res 5 end]
}


proc make_union_case {var hs tp} {
  set res "case $var of "
  foreach v [lrange $tp 3 end] e $hs {
     set n ""
     foreach w [lindex $v 0] {
        set n "$n$w: "
     }
     set res "$res$n[lindex $e 1], "
  }
  set res "([string range $res 0 [expr [string length $res]-3]] end)"
  return $res
}


proc make_query {o path copy_num env scan_partition} {
  global object debug
  if {$debug} {
     puts "### $o"
  }
  set vars {}
  foreach a $object($o,accumulators) {
     lappend vars [new_name]
  }
  set old_env $env
  lappend env [list $object($o,name) $vars $copy_num]
  set heads {}
  set iterations {}
  set predicates {}
  set order {}
  set groupby {}
  set having {}
  set has_groupby 0
  if {[find "partition" $object($o,scope)]} {
     set has_groupby 1
  }
  set unionp 0
  if {[lindex $object($o,otype) 0] == "union"} {
     set unionp 1
  }
  foreach k $object($o,scope) {
     set c $o+[lindex $k 0]
     set struct_name [subst $object($c,name) "'" "_"]
     if {$scan_partition} {
	set v [compile_path "[lindex $k 1]+partition+[lindex $k 0]" [concat $env {{partition PARTITION 0}}] $has_groupby]
     } else {
	set v [compile_path "[lindex $k 1]+[lindex $k 0]" $env $has_groupby]
     }
     if {![winfo exists .c.$c]} {
	set rs [list "" [list [list $struct_name $v]] {} {} {} {} {}]
     } else {
	set rs [make_query $c "[lindex $k 1]+[lindex $k 0]" [lindex $k 2] $env $has_groupby]
     }
     set a [lindex $rs 0]
     set hs [lindex $rs 1]
     set is [lindex $rs 2]
     set ps [lindex $rs 3]
     set os [lindex $rs 4]
     set gs [lindex $rs 5]
     set vs [lindex $rs 6]
     set preds {}
     if {$a == "each"} {
	if {[llength $hs] == 1} {
	   set head [list [list $struct_name [lindex [lindex $hs 0] 1]]]
	} else {
	   set head [list [list $struct_name "struct( [Join $hs {: } { , }] )"]]
	}
	set iterations [concat $iterations $is]
	set preds $ps
	set groupby [concat $groupby $gs]
	set order [concat $order $os]
	set having [concat $having $vs]
	if {$object($c,group)} {
	   set groupby [concat $groupby $head]
	}
     } elseif {$object($c,group)} {
	lappend groupby [list $struct_name [make_oql $rs]]
	set head [list [list $struct_name $object($c,name)]]
     } elseif {$is == {}} {
	if {$a != "some" && $a != "all" && $ps != {} && $object($c,print) \
			 && [type_of $c] == "bool"} {
	   set head [list [list $struct_name "[join $ps " and "] "]]
	} else {
	   set preds [concat $preds $ps]
	   if {$hs == {}} {
	      set head [list [list $struct_name $object($c,name)]]
	   } elseif {$object($o,accumulators) == {}} {
	      set head {}
	      foreach h $hs {
	         set head [concat $head [list [list $object($o,name)\_[lindex $h 0] [lindex $h 1]]]]
	      }
	   } else {
	      set head $hs
	   }
	}
     } else {
	set head [list [list $struct_name [make_oql $rs]]]
     }
     if {$object($c,sort)} {
	lappend order [lindex [lindex $head 0] 1]
     }
     if {$unionp || $object($c,print)} {
	set heads [concat $heads $head]
     } elseif {$a != "each" && [type_of $c] == "bool" \
		 && $v != [lindex [lindex $head 0] 1]} {
	lappend preds [lindex [lindex $head 0] 1]
     }
     if {[info exists object($c,join)] && [valid_path $c $object($c,join)] == 2} {
	set join $object($c,join)
     } else {
	set join ""
     }
     if {$join != ""} {
	lappend preds [list $v $object($c,join_cond) [compile_path $object($c,join) $env $has_groupby]]
     }
     foreach p $object($c,joins) {
	if {[valid_path $p $c] == 1 && $p != $join} {
	   lappend preds [list [compile_path $p $env $has_groupby] $object($p,join_cond) $v]
	}
     }
     if {$object($c,group)} {
	set having [concat $having $preds]
     } else {
	set predicates [concat $predicates $preds]
     }
  }
  set p [compile_path $path $env $has_groupby]
  if {[info exists object($o,head)] && $object($o,head) != ""} {
     lappend heads [list "head" [subst $object($o,head) "\$" $p]]
  }
  if {[info exists object($o,condition)] && $object($o,condition) != ""} {
     if {$object($o,print)} {
	set heads [list [list $object($o,name) [make_predicate [subst $object($o,condition) "\$" $p] $p]]]
     } else {
	lappend predicates [make_predicate [subst $object($o,condition) "\$" $p] $p]
     }
  }
  set var [compile_path $path $old_env $has_groupby]
  if {$unionp} {
     set heads [list [list X [make_union_case $var $heads $object($o,otype)]]]
  }
  set acc ""
  foreach a $object($o,accumulators) X $vars {
     set acc $a
     if {$a == "each"} {
	set iterations [concat [list [list $X $var]] $iterations]
	set var $X
     } else {
	if {$heads == {}} {
	   set heads [list [list $object($o,name) $X]]
	}
	set q [make_oql [list $a $heads [concat [list [list $X $var]] $iterations] \
				$predicates $order $groupby $having]]
	set heads [list [list $object($o,name) $q]]
	set iterations {}
	set predicates {}
     }
  }
  if {$heads == {}} {
     lappend heads [list $object($o,name) $var]
  }
  if {$debug} {
     puts "#-> [list $acc $heads $iterations $predicates $order $groupby $having]"
  }
  return [list $acc $heads $iterations $predicates $order $groupby $having]
}


proc evaluate_oql {} {
  global print_level option
  set opts ""
  for {set x 1} {$x<6} {incr x} {
      set opts "$opts$option($x)"
  }
  catch {
      begin
      set obj [query [.c.oql.up.text get 1.0 end]]
      .c.oql.result.text delete 1.0 end
      set line [toString $obj]
      set count 0
      .c.oql.result.text insert end $line
      .c.oql.result.text insert end "\n"
  } errormsg
  if {$errormsg != "" && [lindex $errormsg 0] != 0} {
      .c.oql.result.text delete 1.0 end
      set count 0
      .c.oql.result.text insert end $errormsg
      .c.oql.result.text insert end "\n"
  }
  commit
}


proc clear_oql {} {
  .c.oql.up.text delete 1.0 end
  .c.oql.result.text delete 1.0 end
}


proc set_print_level { level } {
  global print_level
  set print_level $level
  .c.oql.down.level configure -text "level=$level"
}


proc print_oql {} {
  global tfont print_level option
  if {[winfo exists .c.oql]} {
     clear_oql
  } else {
     toplevel .c.oql
     wm title .c.oql "OQL query"
     frame .c.oql.up
     pack .c.oql.up -side top -expand 0 -fill x
     text .c.oql.up.text -height 8 -width 50 -relief raised -bd 2 -bg white -font $tfont \
		-yscrollcommand ".c.oql.up.scroll set"
     scrollbar .c.oql.up.scroll -command ".c.oql.up.text yview" -orient vertical -bg wheat
     pack .c.oql.up.text -side left -expand 1 -fill x
     pack .c.oql.up.scroll -side right -fill y
     frame .c.oql.down -bg gray
     pack .c.oql.down -fill x -side top
     button .c.oql.down.show -text "generate OQL" -width 10 -command "print_oql" \
		-bg LightSkyBlue -activebackground LightSkyBlue
     button .c.oql.down.save -text "evaluate" -width 6 -command "evaluate_oql" \
		-bg LightSkyBlue -activebackground LightSkyBlue
     button .c.oql.down.clear -text "clear" -width 3 -command "clear_oql" \
		-bg LightSkyBlue -activebackground LightSkyBlue
     set option(1) 0
     set option(2) 0
     set option(3) 1
     set option(4) 0
     set option(5) 0
     set print_level 2
      if {0} {
     menubutton .c.oql.down.menu -text options -menu .c.oql.down.menu.choices -bg gray -activebackground gray -width 10
     menu .c.oql.down.menu.choices -bg gray -activebackground gray -tearoff 0
     .c.oql.down.menu.choices add checkbutton -label "print query optimization steps" -variable option(1)
     .c.oql.down.menu.choices add checkbutton -label "print usage statistics" -variable option(2)
     .c.oql.down.menu.choices add checkbutton -label "perform query unnesting (query decorrelation)" -variable option(3)
     .c.oql.down.menu.choices add checkbutton -label "materialize path expressions into pointer joins" -variable option(4)
     .c.oql.down.menu.choices add checkbutton -label "transform 1:N unnests into pointer joins" -variable option(5)
     menubutton .c.oql.down.level -text "level=$print_level" -menu .c.oql.down.level.choices -bg gray -activebackground gray
     menu .c.oql.down.level.choices -bg gray -activebackground gray -tearoff 0
     .c.oql.down.level.choices add command -label 0 -command "set_print_level 0"
     .c.oql.down.level.choices add command -label 1 -command "set_print_level 1"
     .c.oql.down.level.choices add command -label 2 -command "set_print_level 2"
     .c.oql.down.level.choices add command -label 3 -command "set_print_level 3"
     .c.oql.down.level.choices add command -label 4 -command "set_print_level 4"
     pack .c.oql.down.show .c.oql.down.save .c.oql.down.clear .c.oql.down.menu .c.oql.down.level -padx 15 -side left -fill x
      }
     pack .c.oql.down.show .c.oql.down.save .c.oql.down.clear -padx 15 -side left -fill x
     frame .c.oql.result
     pack .c.oql.result -side left -expand 1 -fill both -side top
     text .c.oql.result.text -height 15 -width 50 -wrap none -relief raised -bd 2 -bg white -font $tfont \
		-yscrollcommand ".c.oql.result.yscroll set" \
		-xscrollcommand ".c.oql.result.xscroll set"
     scrollbar .c.oql.result.xscroll -command ".c.oql.result.text xview" -orient horizontal -bg wheat
     scrollbar .c.oql.result.yscroll -command ".c.oql.result.text yview" -orient vertical -bg wheat
     pack .c.oql.result.xscroll -side top -fill x
     pack .c.oql.result.text -side left -expand 1 -fill both
     pack .c.oql.result.yscroll -side right -fill y -after .c.oql.result.text
     update
     update idletasks
     bind .c.oql.down.show <Any-Leave> {
	clear_message
     }
     bind .c.oql.down.show <Any-Enter> {
	message "Display the OQL query of the current VOODOO query"
     }
     bind .c.oql.down.save <Any-Leave> {
	clear_message
     }
     bind .c.oql.down.save <Any-Enter> {
	message "Evaluate the OQL query using a remote interpreter"
     }
     bind .c.oql.down.clear <Any-Leave> {
	clear_message
     }
     bind .c.oql.down.clear <Any-Enter> {
	message "Clear the OQL and the result window"
     }
      if {0} {
     bind .c.oql.down.menu <Any-Leave> {
	clear_message
     }
     bind .c.oql.down.menu <Any-Enter> {
	message "select interpreter options"
     }
     bind .c.oql.down.level <Any-Leave> {
	clear_message
     }
     bind .c.oql.down.level <Any-Enter> {
	message "select print level"
     }
      }
  }
  set oql [subst [make_oql [make_query root root 0 {} 0]] "root." ""]
  if {$oql == "root"} {
     set oql "struct()"
  }
  .c.oql.up.text insert end "$oql"
  .c.oql.up.text insert end "\n"
}


proc help {} {
  voodoo_error "Please read the VOODOO documentation available at:\nhttp://lambda.uta.edu/ldb/doc/visual.html"
}


proc login {} {
  global server schema username password
  wm title . "VOODOO Setup"
  frame .a
  pack .a
  frame .a.left
  frame .a.right
  pack .a.left .a.right -side left
  label .a.left.l1 -text "Host" -height 1 -width 13 -justify left
  label .a.left.l2 -text "Username" -height 1 -width 13 -justify left
  label .a.left.l3 -text "Password" -height 1 -width 13 -justify left
  label .a.left.l4 -text "Schema" -height 1 -width 13 -justify left
  entry .a.right.t1 -width 20 -relief raised -bd 2 -bg white -textvariable server
  entry .a.right.t2 -width 20 -relief raised -bd 2 -bg white -textvariable username
  entry .a.right.t3 -width 20 -relief raised -bd 2 -bg white -textvariable password
  set schema ""
  entry .a.right.t4 -width 20 -relief raised -bd 2 -bg white -textvariable schema
  pack .a.left.l1 .a.left.l2 .a.left.l3 .a.left.l4
  pack .a.right.t1 .a.right.t2 .a.right.t3 .a.right.t4
  frame .b
  pack .b -side bottom
  button .b.a -text "OK" -command "destroy .a .b"
  button .b.b -text "Cancel" -command "exit"
  pack .b.a .b.b -side left -padx 20
  tkwait window .b
}


proc simple_login {} {
  global server schema username password
  wm title . "VOODOO Setup"
  frame .a
  pack .a
  frame .a.left
  frame .a.right
  pack .a.left .a.right -side left
  label .a.left.l4 -text "Schema" -height 1 -width 13 -justify left
  set schema ""
  entry .a.right.t4 -width 20 -relief raised -bd 2 -bg white -textvariable schema
  pack .a.left.l4
  pack .a.right.t4
  frame .b
  pack .b -side bottom
  button .b.a -text "OK" -command "destroy .a .b"
  button .b.b -text "Cancel" -command "exit"
  pack .b.a .b.b -side left -padx 20
  tkwait window .b
}


proc quit {} {
  close
  exit
}


proc main {} {
  global argv object window_object oql_frame last_join message bfont server port_number classes schema username password
  set repl 1
  while {"$repl" != ""} {
    catch {
	set username admin
	set password pass
	set schema [lindex $argv 0]
	if {$schema == ""} {
	    simple_login
	}
	open $username $password $schema
	begin
	set classes [lindex [tclschema] 0]
	commit
    } repl
    if {"$repl" != ""} {
       voodoo_error "*** Error: $repl"
    } elseif {$classes == {}} {
       voodoo_error "*** Error: schema $schema doesn't exist on server $server:$port_number"
       set repl 1
    } elseif {[string length "$classes"] < 20} {
       voodoo_error "*** Error: server $server:$port_number is already dedicated to serving schema $classes"
       set repl 1
    }
  }
  wm title . "VOODOO: Visual Query Formulation for ODMG OQL"
  wm minsize . 500 500
  wm geometry . +0+0
  canvas .c -width 26c -height 19c -bg wheat -scrollregion {0 0 100c 50c} \
		-xscrollcommand ".xscroll set" -yscrollcommand ".yscroll set"
  scrollbar .xscroll -command ".c xview" -orient horizontal -bg wheat
  scrollbar .yscroll -command ".c yview" -orient vertical -bg wheat
  set message " "
  label .message -textvariable message -justify left -relief sunken -bg wheat
  pack .message -side bottom -fill x
  pack .xscroll -side top -fill x
  pack .c -side left -expand 1 -fill both
  pack .yscroll -side right -fill y -after .c
  menu .c.m -bg PaleTurquoise1 -bd 2 -activebackground turquoise -tearoff 0
  #.c.m add command -label "Help" -command "help"
  .c.m add command -label "Generate OQL" -command "print_oql"
  .c.m add command -label "Quit" -command "quit"
  set w .c.root
  set object(root,name) "root"
  set object(root,otype) "struct()"
  set object(root,gtype) "struct()"
  set object(root,scope) {}
  set object(root,X) 0
  set object(root,Y) 300
  set object(root,accumulators) {}
  frame $w -bg turquoise -bd 4 -relief raised
  set window_object($w) root
  set object(root,window) [.c create window 0 300 -anchor nw -window $w]
  label $w.root -text "Persistent Root" -bg turquoise -font $bfont
  pack $w.root
  frame $w.head -bg turquoise
  pack $w.head
  set last_join {}
  .c move $w 100 100
  update
  update idletasks
  display_struct root [persistent_root]
  print_oql

  bind $w.root <B1-Motion> {
    global window_object object no_message
    set no_message 1
    set win [winfo parent %W]
    raise $win
    set o $window_object($win)
    bind_template_motion $o %x %y 0
  }
  bind $w.root <B2-Motion> {
    global window_object object no_message
    set no_message 1
    set win [winfo parent %W]
    raise $win
    set o $window_object($win)
    bind_template_motion $o %x %y 1
  }
  bind $w.root <B1-ButtonRelease> {
    global no_message
    set no_message 0
  }
  bind $w.root <B2-ButtonRelease> {
    global no_message
    set no_message 0
  }
  bind .c <Button-1> {
    tk_popup .c.m %X %Y
    focus .c.m
  }
  bind $w <Button-1> {
    raise %W
  }
  bind $w.root <Button-1> {
    raise [winfo parent %W]
  }
  bind $w.root <Any-Enter> {
    bind_type_label %W
    message "left-drag: move template,   middle-drag: move template and its children"
  }
  bind $w.root <Any-Leave> {
    clear_label
    clear_message
  }
  bind $w <Any-Enter> {
    %W.root configure -bg PaleTurquoise1
  }
  bind $w <Any-Leave> {
    %W.root configure -bg turquoise
  }
  bind .c <Any-Enter> {
    message "click the left mouse button for a menu"
  }
  bind .c <Any-Leave> {
    clear_message
  }
}


main
