/*********************************************************************************
 *
 *  Functions to manipulate ODMG ODL 2.0 database catalog
 *  Copyright (c) 1999-2003 by Leonidas Fegaras, the University of Texas at
 *     Arlington. All rights reserved.
 *  Programmer: Leonidas Fegaras
 *  Date: 9/28/98
 *
 ********************************************************************************/


#ifndef __odl_h__
#define __odl_h__


#include "gc_cpp.h"
#include "ldb_sm.h"
#include "basic.h"


#if defined(linux) || defined(__linux__)
const bool LDB_LINUX = true;
#else
const bool LDB_LINUX = false;
#endif


enum type_tag { error_dcl, class_dcl, interface_dcl, forward_dcl, extent_dcl,
		typedef_dcl, constant_dcl, variable_dcl, index_dcl,
		attribute_dcl, relationship_dcl, method_dcl,
		exception_dcl, struct_dcl, union_dcl, enum_dcl,
		define_dcl, class_file_dcl, module_dcl, counter_dcl, compiled_method_dcl };


/* An entry in the user catalog */
class Catalog {
public:
  raw_string		name;		// the pathname of the declaration item
  type_tag		tag;            // the kind of the declaration
  raw_string		binding;	// the actual declaration in Expr form
					//    (laid out as a byte string)
  raw_string            module;         // module name
};


enum data_entry_tag { extent_entry, index_entry, temp_extent_entry };


/* A user data entry, such as a class extent or a B+-tree index */
class DataRepository {
public:
  data_entry_tag        tag;
  raw_string		name;   	// extent or index pathname
  raw_string		class_name;	// pathname of the owner class
  raw_string            module;         // module name
  size_t		cardinality;	// extent or index cardinality
  size_t		size;		// extent or index size in bytes
  oid_t                 oid;            // location in database               
  bool                  uniquep;        // true for primary index (unique keys)
};


/* Info about a user */
class UserInfo {
public:
  raw_string   name;
  raw_string   password;
  oid_t        catalog;
  oid_t        catalog_index;
  oid_t        user_database;
  oid_t        user_database_index;
};


class Path: public gc {
private:
  list<String>* context;
public:
  Path ();
  void new_context ( String name );
  void exit_context ();
  String current () const { return context->hd; };
  String local ( String name );
  String clocal ( String name );
  list<String>* local_in_each_module ( String name );
};

void odmg_error (const char* s, Expr e);

void odmg_error (const char* s, String e);

void odmg_error (const char* s);

Ref<UserInfo> create_user ( sm_string username, sm_string password );

Ref<UserInfo> login ( sm_string username, sm_string password );

void initialize_current_module ();

void set_current_module ( String name );

String get_local_in_path ( String name );

void print_all_declarations ();

list< Ref<Catalog> >* get_declarations ( String dname );

Ref<Catalog> get_declaration ( String dname );

Expr get_declaration_binding ( Ref<Catalog> dcl );

Expr get_declaration_binding ( String dname );

Expr get_declaration_binding ( String dname, String cname );

Ref<Catalog> store_declaration ( String name, const type_tag tag, Expr binding,
			         bool duplicate_error = true );

Ref<DataRepository> get_data_entry ( String pathname );

Expr find_in_module ( Expr nm );

Ref<DataRepository> new_extent ( String pathname, String classname );

Ref<DataRepository> new_index ( String pathname, String classname, bool uniquep = true );

Expr find_attribute_type ( String attr, String cname, String &iname );

bool classp ( String name );

Expr class_type ( String cname );

Expr extent_type ( String ename );

bool extentp ( Expr x );

bool constantp ( Expr x );

Expr get_class_extent ( Expr x );

list<Expr>* class_keys ( Expr e );

list<String>* primary_key ( String s );

bool subclass ( String x, String y );

list<String>* class_superclasses ( String class_name );

list<String>* class_subclasses ( String class_name );

list<String>* module_classes ();

String new_secondary_index ( Expr e );

// colect database statistics (eg. cardinalities) and store them in the system catalog
void collect_statistics ();

void tcl_print_module ( ostream &sout );

list<Expr>* get_union_tags ( Expr type );

#endif
