/********************************************************************************
*
* The optimizer generator (OPTGEN) parser that parses OPTL programs
*
* Copyright (c) 1999-2003 by Leonidas Fegaras, the University of Texas at
*     Arlington. All rights reserved.
* Programmer: Leonidas Fegaras, UTA
* Date: 9/10/97
*
*********************************************************************************/
%{
#include <stream.h>
#include <fstream.h>
#include <strstream.h>
#include <string.h>
#include "external.h"


bool put_gen_lines = true;

ofstream fstring;

static bool file_IO = false;

static int lineno = 1;

static bool in_rule = false;

static list<String>* all_strings = new list<String>;

const Expr _variable = variable("variable");
const Expr _EMPTY_NAME_ = variable("");
const Expr _function = variable("function");
const Expr _Cons = variable("Cons");
const Expr _Nil = variable("Nil");
const Expr _integer = variable("integer");
const Expr _real = variable("real");
const Expr _estring = variable("estring");
const Expr _append = variable("Append");


class lstream: public ofstream {
  char buffer[1000];
  bool stored_output;
  ostrstream* stored_stream;
  ofstream*  output_stream;
public:
  lstream ( const char* file_name ) {
       stored_output = false;
       stored_stream = new (GC) ostrstream(buffer,1000);
       output_stream = new (GC) ofstream(file_name);
  };
  void stored () { stored_output = true; };
  void output () { stored_output = false; };
  const char* content () {
    char* s = stored_stream->str();
    s[stored_stream->pcount()-1]='\0';
    return s;
  };
  void reset () { stored_stream = new (GC) ostrstream(buffer,1000); };
  void flush () { output_stream->flush(); };
  template <class T>
       friend lstream& operator<< ( lstream& o, const T &value ) {
          if (o.stored_output)
	     *o.stored_stream << value;
	  else *o.output_stream << value;
	  return o;
       };
};


static lstream* fout;


class continuation: public gc {
  list<String>* value;
public:
  continuation () { value = (new list<String>)->cons(new string("")); };
  void remember ( const char* str ) { string s = string(str) + *value->hd; value->hd = new string(s); };
  void remember () { remember("}"); };
  void close () { String s = value->hd; value = value->tl; *fout << s; };
  void open () { value = value->cons(new string("")); };
};

static continuation* cont = new continuation;


void yyerror (char*);


class stop_symbols: public gc {
  list<list<int>*>* stack;
public:
  stop_symbols() { stack = new list<list<int>*>; }
  void push() { stack = stack->cons(new list<int>); };
  void push(char c) { stack = stack->cons((new list<int>)->cons(c)); };
  void push(int i) { stack = stack->cons((new list<int>)->cons(i)); };
  void push(char c, int n) { stack = stack->cons((new list<int>)->cons(c)->cons(n)); };
  void pop() { if (stack->consp()) stack = stack->tl;
               else yyerror("Unbalanced parentheses"); };
  list<int>* top() { if (stack->consp()) return stack->hd;
                     else yyerror("Unbalanced parentheses"); };
  void print() { for(list<list<int>*>* r = stack; r->consp(); r=r->tl) cerr << r->hd << " "; cerr << endl; };
};

static stop_symbols* stopsyms = new stop_symbols;


int fsyms ( int sym ) {
  if (stopsyms->top()->nullp())
     return sym;
  switch (sym)
  { case '(' : stopsyms->push(')'); break;
    case '{' : stopsyms->push('}'); break;
    case '[' : stopsyms->push(']'); break;
    case ')' : stopsyms->pop(); break;
    case '}' : stopsyms->pop(); break;
    case ']' : stopsyms->pop(); break;
    default: for(list<int>* r = stopsyms->top(); r->consp(); r=r->tl) 
               if (sym == r->hd) return sym;
  };
  return SPECIAL;
};

#include "optgen.yy.c"

void yyerror (char *s) {
   cerr << s << " (line: " << lineno << ")" << " Token: " << yytext << endl;
   exit(-1);
};


static int rule_number;
static Expr rule_pattern;
static list<String>* inherited = new list<String>;
static list<String>* synthesized = new list<String>;
static int logical_module_size = 0;
list<int>* logical_modules = new list<int>;


Expr reify ( Expr e ) {
  if (e->functionp())
  { list<Expr> *params = e->info.Function.Parameters;
    if (e->info.Function.Name->eq(variable("VAR")) && (params->consp()))
       if (params->hd->variablep())
	  return variable(params->hd->info.Variable);
    Expr ps = _Nil;
    for(list<Expr>* r = params->reverse(); r->consp(); r=r->tl)
       if (r->hd->functionp() && r->hd->info.Function.Name->eq(variable("VARN"))
           && r->hd->arguments()->consp()
           && r->hd->arguments()->hd->variablep())
	 ps = function(_append,Cons(variable(r->hd->arguments()->hd->info.Variable),Cons(ps,Nil)));
       else ps = function(_Cons,Nil->cons(ps)->cons(reify(r->hd)));
    return function(_function,Nil->cons(ps)->cons(reify(e->info.Function.Name)));
  } else if (e->variablep())
         return function(_variable,Nil->cons(estring(e->info.Variable)));
    else if (e->integerp())
         return function(_integer,Nil->cons(integer(e->info.Integer)));
    else if (e->realp())
         return function(_real,Nil->cons(real(e->info.Real)));
    else if (e->stringp())
         return function(_estring,Nil->cons(estring(e->info.Estring)));
  return variable("error");
};


void compile_pattern (expr *pattern, String exp, bool casep ) {
   if (pattern->variablep())
      if (file_IO)
	 *fout << " if (" << exp << "->variablep())\n if (" << exp << "->eq(_" << pattern << "_))\n";
      else *fout << " if (" << exp << "->variablep())\n if (" << exp << "->info.Variable->eq(\"" << pattern << "\"))\n";
   else if (pattern->integerp())
      *fout << " if (" << exp << "->integerp())\n if (" << exp << "->info.Integer == " << pattern << ")\n";
   else if (pattern->realp())
      *fout << " if (" << exp << "->realp())\n if (" << exp << "->info.Real == " << pattern << ")\n";
   else if (pattern->stringp())
      *fout << " if (" << exp << "->realp())\n if (" << exp << "->info.Estring->eq(" << pattern << "))\n";
   else if (pattern->functionp())
      { list<Expr>* params = pattern->info.Function.Parameters;
        if (pattern->info.Function.Name->eq(variable("VAR")) && (params->consp()))
        {  if (params->hd->variablep())
	   { *fout << " { " << params->hd << " = " << exp << ";\n";
	     cont->remember();
           }
	}
        else if (!casep && pattern->info.Function.Name->eq(variable("RVAR")) && (params->consp()))
        {  if (params->tl->hd->variablep())
              { *fout << " { " << params->tl->hd << " = " << exp << ";\n";
		cont->remember();
              }
	}
        else if (casep && pattern->info.Function.Name->eq(variable("RVAR")) && (params->consp()))
        { if (params->tl->hd->variablep())
         { String nm = params->tl->hd->info.Variable;
	   String nv = new_name();
	   String rv = new_name();
           *fout << " { Expr _" << nm << "; " << nm << " = " << exp << "; "
		<< " String " << rv << " = new_name();\n Expr " << nv << ";\n list<Expr>* stack_" << nv << " = Cons(" << nm << ",Nil);\n"
		<< " while (!FOUND && stack_" << nv << "->consp()) {\n"
		<< " " << nv << " = stack_" << nv << "->hd;\n stack_" << nv << " = stack_" << nv << "->tl;\n"
	        << " if (" << nv << "->functionp()) stack_" << nv << " = " << nv << "->arguments()->append(stack_" << nv << ");\n";
	   compile_pattern(pattern->arguments()->hd,nv,casep);
	   *fout	<< "{ _" << nm << " = function(variable(\"bind\"),Cons(variable(" << rv << "),Cons(subst_expr("
		<< reify(params->hd)
		<< ",variable(" << rv << ")," << nm << "),Nil)));\n";
	   cont->remember("}}};\n");
	 }
        }
	else {
	if (pattern->info.Function.Name->functionp()
           && pattern->info.Function.Name->info.Function.Name->eq(variable("VAR"))
	   && pattern->info.Function.Name->info.Function.Parameters->consp())
        { *fout << " if (" << exp << "->functionp())\n"
	       << " { " << pattern->info.Function.Name->info.Function.Parameters->hd
	       << " = " << exp << "->info.Function.Name;\n";
	  cont->remember();
	}
        else *fout << " if (" << exp << "->functionp())\n"
		  << " if (" << exp << "->info.Function.Name->eq(_"
		  << pattern->info.Function.Name
		  << "_))\n";
        String cexp = new string(*exp + "->info.Function.Parameters");
        list<Expr> *r = params;
        for( ; r->consp(); r=r->tl )
           { if (r->hd->functionp())
		if (r->hd->info.Function.Name->eq(variable("VARN"))
                    && (r->hd->info.Function.Parameters->consp()))
                   if (r->hd->info.Function.Parameters->tl->nullp())
		   { String nm = new_name();
		     String vm = r->hd->info.Function.Parameters->hd->info.Variable;
		     if (r->tl->nullp())
		     { *fout << " { " << vm << " = " << cexp << ";\n";
		       cont->remember();
		       break;
                     };
		     *fout << " { " << vm << " = Nil;\n for(list<Expr>* " << nm << " = " << cexp << "; "
		          << (in_rule?"":"!FOUND && ") << nm << "->consp(); " << vm << "=(" << (in_rule?"true":"!FOUND")
			  << " ? " << vm << "->append(Cons(" << nm << "->hd,Nil)) : " << vm << "), " << nm << "=" << nm << "->tl)\n";
		     cont->remember();
		     cexp = nm;
		     continue;
		   };
             *fout << " if ((" << cexp << ")->consp())\n";
             compile_pattern(r->hd,new string(*cexp + "->hd"),casep);
	     string s = *cexp + "->tl";
             cexp = new string(s);
          };
        if (r->nullp()) *fout << " if (" << cexp << "->nullp())\n";
      }
    }
};


void var_decl ( Expr e, const char* type_name ) {
  if (e->functionp())
  { list<Expr>* params = e->info.Function.Parameters;
    if (e->info.Function.Name->functionp()
	&& e->info.Function.Name->info.Function.Name->eq(variable("VAR"))
	&& e->info.Function.Name->info.Function.Parameters->consp())
       { if (e->info.Function.Name->info.Function.Parameters->hd->variablep())
            *fout << type_name << " " << e->info.Function.Name->info.Function.Parameters->hd << ";\n";
       };
    if (e->info.Function.Name->eq(variable("VAR")) && (params->consp()))
       { if (params->hd->variablep())
            *fout << type_name << " " << params->hd << ";\n";
       }
    else if (e->info.Function.Name->eq(variable("RVAR")) && (params->consp()))
       { if (params->tl->hd->variablep())
            *fout << type_name << " " << params->tl->hd << ";\n";
	 var_decl(params->hd,type_name);
       }
    else if (e->info.Function.Name->eq(variable("VARN")) && (params->consp()))
       { if (params->hd->variablep() && (params->tl->nullp()))
            *fout << "list<" << type_name << ">* " << params->hd << ";\n";
       }
    else for( list<Expr>* r = params; r->consp(); r=r->tl ) var_decl(r->hd,type_name);
  };
};


void rewrite_vars ( Expr e, String rname) {
  if (e->functionp())
  { list<Expr> *params = e->info.Function.Parameters;
    if (e->info.Function.Name->eq(variable("VAR")) && (params->consp()))
       { if (params->hd->variablep())
         { String nm = params->hd->info.Variable;
           *fout << " for(list<plan*>* _" << nm << " = " << rname << "_REWRITE(" << nm << ",";
	   if (params->tl->nullp()) *fout << "ATRBS";
           else if (params->tl->hd->functionp())
		{ *fout << "inherit(";
		  for(list<Expr>* r = params->tl->hd->info.Function.Parameters; r->consp(); r=r->tl)
		  {  if (r->hd->functionp())
			if (r->hd->info.Function.Name->eq(variable("#")))
			   *fout << r->hd->info.Function.Parameters->hd;
			else *fout << "ATRBS." << r->hd;
		     else *fout << "ATRBS." << r->hd;
		     if (r->tl->consp()) *fout << ",";
		  };
		  *fout << ")";
		} else *fout << "ATRBS";
           *fout << ",LEVEL+1); _" << nm << "->consp(); _" << nm << "=_" << nm << "->tl) {\n"
		<< " Expr " << nm << " = _" << nm << "->hd->expression;\n";
	   cont->remember();
	 }
       }
    else if (e->info.Function.Name->eq(variable("RVAR")) && (params->consp()))
       { if (params->tl->hd->variablep())
         { String nm = params->tl->hd->info.Variable;
           *fout << " for(list<plan*>* _" << nm << " = " << rname << "_REWRITE(" << nm << ",";
	   if (params->tl->tl->nullp()) *fout << "ATRBS";
           else if (params->tl->tl->hd->functionp())
		{ *fout << "inherit(";
		  for(list<Expr>* r = params->tl->tl->hd->info.Function.Parameters; r->consp(); r=r->tl)
		  {  if (r->hd->functionp())
			if (r->hd->info.Function.Name->eq(variable("#")))
			   *fout << r->hd->info.Function.Parameters->hd;
			else *fout << "ATRBS." << r->hd;
		     else *fout << "ATRBS." << r->hd;
		     if (r->tl->consp()) *fout << ",";
		  };
		  *fout << ")";
		} else *fout << "ATRBS";
	   String nv = new_name();
	   String rv = new_name();
           *fout << ",LEVEL+1); _" << nm << "->consp(); _" << nm << "=_" << nm << "->tl) {\n"
		<< " bool RFOUND = false; String " << rv << " = new_name();\n Expr " << nv << ";\n"
		<< " list<Expr>* stack_" << nv << " = Cons(_" << nm << "->hd->expression,Nil);\n"
		<< " loop_" << nv << ": if (stack_" << nv << "->nullp()) goto exit_" << nv << ";\n"
		<< " " << nv << " = stack_" << nv << "->hd;\n stack_" << nv << " = stack_" << nv << "->tl;\n";
	   cont->open();
	   compile_pattern(e->arguments()->hd,nv,false);
	   *fout	<< "{ RFOUND = true; _" << nm << " = function(variable(\"bind\"),Cons(variable(" << rv << "),Cons(subst_expr("
		<< reify(params->hd)
		<< ",variable(" << rv << "),_" << nm << "->hd->expression),Nil)));\n";
	   cont->close();
	   *fout << "}; if (RFOUND) {\n";
           string xss = "}};\n if (" + *nv + "->functionp()) stack_" + *nv + " = " + *nv + "->arguments()->append(stack_"
		        + *nv + ");\n goto loop_" + *nv + ";\n exit_" + *nv + ": ;\n";
	   cont->remember(xss.content());
	 }
       }
    else if (e->info.Function.Name->eq(variable("VARN")) && (params->consp()))
       { if (params->hd->variablep())
         { String nm = params->hd->info.Variable;
	   *fout << " for(list<plan*>* _" << nm << " = default_rewrite(function(_EMPTY_NAME_," << nm << "),"
		<< "ATRBS,LEVEL+1," << rname << "_REWRITE); _" << nm << "->consp(); _" << nm << "=_" << nm << "->tl) {\n"
		<< " list<Expr>* " << nm << " = _" << nm << "->hd->expression->info.Function.Parameters;\n";
	   cont->remember();
	 }
       }
    else for( list<Expr> *r = params; r->consp(); r=r->tl )
	    rewrite_vars(r->hd,rname);
  };
};


Expr change_attributes ( binding<Expr>* bl ) {
  list<Expr>* params = Nil;
  for(list<String>* r = inherited; r->consp(); r = r->tl)
  { params = params->cons(variable(r->hd));
    for(Names s = bl->names(); s->consp(); s = s->tl)
       if (s->hd->eq(r->hd))
          params = params->tl->cons(function(variable("#"),Nil->cons(bl->find(s->hd))));
  };
  return function(variable("ATRBS"),params->reverse());
};


void prelude (String name) {
  *fout << "list<plan*>* " << name << "_REWRITE ( Expr EXPR, inherited ATRBS, int LEVEL ) {\n"
       << "memo* MEMO = memoized(" << name << "_hash_table,EXPR,ATRBS);\n"
       << "if (MEMO != 0) {\n"
       << "   /* if (trace_rules) trace_output(LEVEL,0,MEMO->plans); */\n"
       << "   return MEMO->plans;\n"
       << "};\n"
       << "list<plan*>* PLANS = EMPTY_PLANS;\n"
       << "list<plan*>* NEW_PLANS = EMPTY_PLANS;\n"
       << "MEMO = memoize_empty(" << name << "_hash_table,EXPR,ATRBS);\n"
       << "if (trace_rules && EXPR->functionp()) trace_input(LEVEL,0,EXPR,ATRBS);";
};

%}
%union {
   String				Tstring;
   list<String>*			Tstrings;
   int					Tint;
   bool					Tbool;
   Expr					Texpression;
   list<Expr>*				Tparameters;
   pair<String,Expr>*			Tbinding;
   binding<Expr>*			Tbindings;
}
%type <Tstring> name rname
%type <Texpression> pattern exp rest variable patvar prelude cprelude
%type <Tparameters> patterns exps
%type <Tint> condition patrbs
%type <Tbool> becomes pcases pcase pbinds
%type <Tbinding> atrbpat atrbexp
%type <Tbindings> atrbpats attributes atrbexps
%type <Tstrings> names synths
%token ID NUM FLOAT CSTRING SPECIAL META CASE END IMPLIES DOTS IS LOGICAL PHYSICAL PERMUTATION INHERITED SYNTHESIZED COMPUTE LET IN DO FORALL TRACE FAIL
%%
main:	  ccode			{ stopsyms->push(); }
	  optgen
	;
ccode:    ctoken
	| ccode ctoken
	;
ctoken:	  name			{ *fout << $1; }
	| NUM			{ *fout << yytext; }
	| FLOAT			{ *fout << yytext; }
	| CSTRING		{ *fout << yytext; }
	| '^' name		{ *fout << "_" << $2 << "->hd->attributes"; }
	| SPECIAL		{ *fout << yytext; }
	| '#' name		{ *fout << "#" << $2; }
	| META 			{ stopsyms->push(); }
	  exp '>' 		{ *fout << $3; stopsyms->pop(); }
	| CASE			{ stopsyms->push('|');
				  *fout << "{ bool FOUND = false;\n";
				  if (put_gen_lines) *fout << "#line " << lineno << '\n';
				  *fout <<  " Expr EXPR = "; }
	  ccode			{ *fout << ";\n";
				  stopsyms->pop();
				  stopsyms->push(); }
          cases			{ stopsyms->pop(); }
	  END			{ *fout << "\n}";
				  if (put_gen_lines) *fout << "\n#line " << lineno << '\n'; }
	;
case:	  '|'			{ cont->open();
				  *fout << "if (!FOUND) {\n";
				  cont->remember(); }
	  cprelude condition IMPLIES
				{ stopsyms->pop();
				  stopsyms->push('|',END);
  				  if ($4) cont->remember("}};");
				  if (put_gen_lines) *fout << "\n#line " << lineno << '\n'; }
	  ccode			{ *fout << "\n FOUND = true;";
				  cont->close();
				  stopsyms->pop();
				  stopsyms->push(); }
	;
cases:	  case
	| cases case
	;
prelude:  pattern		{ *fout << "{ ";
				  cont->remember();
				  var_decl($1,"Expr");
				  compile_pattern($1,new string("EXPR"),false);
				  *fout << " {\n";
				  cont->remember();
				  $$ = $1; }
	| LET blist IN prelude	{ *fout << "{ ";
				  cont->remember();
				  var_decl($4,"Expr");
				  compile_pattern($4,new string("EXPR"),false);
				  *fout << " {\n";
				  cont->remember();
				  $$ = $4; }
	;
cprelude:  pattern		{ *fout << "{ ";
				  cont->remember();
				  var_decl($1,"Expr");
				  compile_pattern($1,new string("EXPR"),true);
				  *fout << " {\n";
				  cont->remember();
				  $$ = $1; }
	;
name:	  ID			{ $$ = new string(yytext); }
	;
variable: name			{ if (!member(variable($1),all_strings))
				     all_strings = all_strings->cons($1);
				  if (file_IO)
				     $$ = variable(new string("_" + *$1 + "_"));
				  else $$ = function(_variable,Nil->cons(estring($1))); }
	| '`' name		{ $$ = variable($2); }
	| '`' '('		{ fout->reset();
				  fout->stored();
				  stopsyms->push(')'); }
	  ccode			{ char* s = (char*) fout->content();
				  if (s[strlen(s)-1] == ' ')
				  {  int i;
				     for(i = strlen(s)-1; s[i]!=')'; i--)
				        s[i] = '\0';
				     s[i] = '\0';
				  };
				  $$ = function(_EMPTY_NAME_,Nil->cons(variable(new string(s))));
				  fout->output(); }
	| '_'			{ $$ = function(_variable,Nil->cons(estring("_"))); }
	;
exp:	  NUM			{ $$ = function(_integer,Nil->cons(integer(atoi(yytext)))); }
	| FLOAT			{ $$ = function(_real,Nil->cons(real(atof(yytext)))); }
	| CSTRING		{ yytext[strlen(yytext)-1] = '\0';
				  $$ = function(_estring,Nil->cons(estring(new string(yytext+1)))); }
	| variable		{ $$ = $1; }
	| variable '(' ')'	{ $$ = function(_function,Nil->cons(_Nil)->cons($1)); }
	| variable '(' exps ')'
			{ Expr params = _Nil;
			  for(list<Expr>* r = $3; r->consp(); r=r->tl)
			     if (r->hd->functionp() && r->hd->info.Function.Name->eq(variable("*")))
				if (r != $3)
				   params = function(variable("Append"),
						Nil->cons(params)->cons(r->hd->info.Function.Parameters->hd));
				else params = r->hd->info.Function.Parameters->hd;
			     else params = function(_Cons,Nil->cons(params)->cons(r->hd));
			  $$ = function(_function,Nil->cons(params)->cons($1)); }
	| '`' name '[' exp ']'
			{ $$ = function(variable("substitute"),Cons(variable(new string("_" + *$2)),Cons($4,Nil))); }
	| '(' exp ')'		{ $$ = $2; }
	;
exps:	  exp			{ $$ = Nil->cons($1); }
	| rest			{ $$ = Nil->cons(function(variable("*"),Nil->cons($1))); }
	| exps ',' rest		{ $$ = $1->cons(function(variable("*"),Nil->cons($3))); }
	| exps ',' exp		{ $$ = $1->cons($3); }
	;
rest:	  /* empty */		{ $$ = _Nil; }
	| DOTS name		{ $$ = variable($2); }
	| DOTS '('		{ fout->reset();
				  fout->stored();
				  stopsyms->push(')'); }
	  ccode			{ $$ = variable(new string(fout->content()));
				  fout->output(); }
	| DOTS			{ $$ = variable("_"); }
	;
patvar:   name			{ if (!member(variable($1),all_strings))
				     all_strings = all_strings->cons($1);
				  $$ = variable($1); }
	| '`' name		{ $$ = function(variable("VAR"),Nil->cons(variable($2))); }
	| '`' name IS '{' atrbexps '}'
		{ $$ = function(variable("VAR"),Nil->cons(change_attributes($5))->cons(variable($2))); }
	| '`' name '[' pattern ']'
			{ $$ = function(variable("RVAR"),Nil->cons(variable($2))->cons($4)); }
	| '`' name '[' pattern ']' IS '{' atrbexps '}'
			{ $$ = function(variable("RVAR"),Nil->cons(change_attributes($8))->cons(variable($2))->cons($4)); }
	| '_'			{ $$ = function(variable("VAR"),Nil->cons(new_variable())); }
	;
pattern:  NUM			{ $$ = integer(atoi(yytext)); }
	| FLOAT			{ $$ = real(atof(yytext)); }
	| CSTRING		{ yytext[strlen(yytext)-1] = '\0';
				  $$ = estring(new string(yytext+1)); }
	| patvar		{ $$ = $1; }
	| patvar '(' patterns ')'
			{ $$ = function($1,$3->reverse()); }
	| patvar '(' ')'
			{ $$ = function($1,Nil); }
	| '(' pattern ')'	{ $$ = $2; }
	;
patterns: pattern		{ $$ = Nil->cons($1); }
	| DOTS name		{ $$ = Nil->cons(function(variable("VARN"),Nil->cons(variable($2)))); }
	| patterns ',' DOTS name { $$ = $1->cons(function(variable("VARN"),Nil->cons(variable($4)))); }
	| patterns ',' pattern	{ $$ = $1->cons($3); }
	;
optgen:   /* empty */
	| '%' '{'	{ stopsyms->push('%'); }
	  ccode		{ stopsyms->pop();
			  *fout << "#ifndef trace_rules\n#define trace_rules false\n#endif\n"
			       << "#ifndef trace_stages\n#define trace_stages false\n#endif\n"
			       << "#ifndef hash_size\n#define hash_size 1000\n#endif\n"; }
	  '%' TRACE	{ *fout << "#define rules_to_trace "; stopsyms->push('%'); }
	  ccode		{ *fout << '\n'; stopsyms->pop(); }
	  '%' INHERITED		{ *fout << "struct inherited {"; }
	  names			{ for(list<String>* r = $14; r->consp(); r = r->tl)
				     inherited = inherited->cons(r->hd);
				  for(list<String>* r = inherited; r->consp(); r = r->tl)
				     *fout << " Expr " << r->hd << ";";
				  *fout << "\n};\n\ninherited inherit( Expr _" << inherited->hd;
				  for(list<String>* r = inherited->tl; r->consp(); r = r->tl)
				     *fout << ", Expr _" << r->hd;
				  *fout << " ) { inherited atrb; ";
				  for(list<String>* r = inherited; r->consp(); r = r->tl)
				     *fout << "atrb." << r->hd << "=_" << r->hd << "; ";
				  *fout << "return atrb; };\n\n"
				       << "ostream& operator<< ( ostream& s, inherited x ) { return s";
				  for(list<String>* r = inherited; r->consp(); r = r->tl)
				     *fout << " << \"" << r->hd << ": \" << x." << r->hd << " << \",\"";
				  *fout << "; };\n\n"; }
	  '%' SYNTHESIZED	{ *fout << "struct synthesized {\n"; }
	  '{' synths '}' '='	{ stopsyms->push('%');
				  *fout << "\n};\n\nsynthesized default_synthesized = "; }
	  ccode	 { *fout << ";\n\n"
			<< "class plan: public gc {\npublic:\n  Expr expression;\n  synthesized attributes;\n"
			<< "  plan ( Expr e, synthesized s ) { expression = e; attributes = s; };\n};\n\n"
			<< "list<plan*>* EMPTY_PLANS = new list<plan*>;\n\n"
			<< "class memo: public gc {\npublic:\n  Expr expression;\n  inherited attributes;\n"
			<< "  list<plan*>* plans;\n"
			<< "  memo ( Expr e, inherited s, list<plan*>* p )\n"
			<< "       { expression = e; attributes = s; plans = p; };\n};"
			<< "\n\nshort eqatrb ( inherited X, inherited Y ) { return ";
		   for(list<String>* r = inherited; r->consp(); r = r->tl)
		      *fout << "X." << r->hd << "->eq(Y." << r->hd << ") && ";
		   *fout << "true; };\n\n#include \"common.h\"\n\n";
		   stopsyms->pop();
		   synthesized = $20; }
	  logrules '%' PHYSICAL
				{ *fout << "\n\nlist<plan*>* all_plans_ (list<plan*>* x) { return x; };\n"
				       << "static list<plan*>* (*best_plans) (list<plan*>*) = &all_plans_;\n\n";
				  prelude(new string("PHYSICAL")); }
	  prules
				{ *fout << "  if (PLANS->nullp())\n"
				       << "     PLANS = default_rewrite(EXPR,ATRBS,LEVEL,PHYSICAL_REWRITE);\n"
				       << "  PLANS = (*best_plans)(PLANS);\n"
				       << "if (trace_rules && EXPR->functionp()) trace_output(LEVEL,0,PLANS);\n"
				       << "MEMO->plans = PLANS; return PLANS;\n};\n";
				  if (put_gen_lines) *fout << "#line " << lineno << '\n'; }
	  '%' '}'		{ stopsyms->pop();
				  *fout << "\nvoid optimizer_init () {\n"
				       << "  int x[100] = rules_to_trace;\n"
				       << "  for(short i = 0; i<100; i++) rules2trace[i] = x[i];\n"
				       << "  if (trace_rules) { cout << \"Tracing the rules: \";\n  for(short i = 0; x[i]; i++) cout << x[i] << \",\";\n  cout << endl; };\n";
				  for(int i = 1; i<=logical_module_size; i++)
				     *fout << "  hash_table_init(LOGICAL" << i << "_hash_table);\n";
				  *fout << "  hash_table_init(PHYSICAL_hash_table);\n};\n\n";
				  *fout << "list<plan*>* rewrite ( Expr EXPR, inherited ATRBS, bool generate_plans ) {\n"
				       << "  list<plan*>* res = EMPTY_PLANS;\n";
				  for(int i = 1; i<=logical_module_size; i++)
				  {  if (i==1)
				        *fout << "  for(list<plan*>* p" << i << " = LOGICAL" << i << "_REWRITE(EXPR,ATRBS,0);";
				     else *fout << "  for(list<plan*>* p" << i << " = LOGICAL" << i << "_REWRITE(p"
				               << i-1 << "->hd->expression,ATRBS,0);";
				     *fout << " p" << i << "->consp(); p" << i << "=p" << i << "->tl)\n"
					  << "  { if (trace_stages)\n    { cout << \"\\nStage " << i << ": \"; p" << i << "->hd->expression->pretty_print(9,cout); cout << endl; };\n";
			          };
				  *fout << "     if (generate_plans)\n"
				       << "         res = (*best_plans)(res->append(PHYSICAL_REWRITE(p" << logical_module_size << "->hd->expression"
				       << ",ATRBS,0)));\n else res = res->cons(p" << logical_module_size << "->hd);\n";
				  for(int i = 1; i<=logical_module_size; i++) *fout << " };";
				  *fout << "   return res;\n};\n";
				  if (put_gen_lines) *fout << "#line " << lineno << '\n';
				}
	  ccode
	;
logrule:  '%' LOGICAL		{ logical_modules = logical_modules->cons(logical_module_size++);
				  String name = new string("LOGICAL" + string(logical_module_size));
				  *fout << "static list<memo*>* " << name << "_hash_table[hash_size];\n\n";
				  prelude(name); }
	  lrules		{ *fout << " if (PLANS->nullp()) PLANS = default_rewrite(EXPR,"
				       << "ATRBS,LEVEL,LOGICAL" << logical_module_size << "_REWRITE);\n";
				  *fout << "if (trace_rules && EXPR->functionp()) trace_output(LEVEL,0,PLANS);\n"
				       << "MEMO->plans = PLANS; return PLANS;\n};\n";
				  if (put_gen_lines) *fout << "#line " << lineno << '\n'; }
	;
logrules: logrule
	| logrules logrule
	;
names:	  name			{ $$ = (new list<String>)->cons($1); }
	| names ',' name	{ $$ = $1->cons($3); }
	;
synths:	  name name ';'		{ *fout << $1 << " " << $2 << ";";
				  $$ = (new list<String>)->cons($2); }
	| synths name name ';'	{ *fout << $2 << " " << $3 << ";";
				  $$ = $1->cons($3); }
	;
lrule:				{ cont->open(); }
	  patrbs prelude	{ *fout << " if (trace_rules && EXPR->functionp()) trace_input(LEVEL," << rule_number << ",EXPR,ATRBS);\n";
				  in_rule = false;
				  rewrite_vars($3,new string("LOGICAL" + string(logical_module_size)));
				  if (put_gen_lines) *fout << "#line " << lineno << '\n';
				  *fout << "{ \nNEW_PLANS = EMPTY_PLANS;\n";
				  rule_pattern = $3; }
	  condition becomes binds ';'
				{ *fout << " if (trace_rules && EXPR->functionp()) trace_output(LEVEL," << rule_number << ",NEW_PLANS);\n";
				  if ($6) *fout << " MEMO->plans = NEW_PLANS;\n return NEW_PLANS;";
				     else *fout << " PLANS = merge_plans(PLANS,NEW_PLANS);";
				  if ($5) *fout << "\n } else if (trace_rules && EXPR->functionp()) trace(LEVEL," << rule_number << ",\"condition failed\"); };";
				  cont->close();
				  *fout << "} }; exit_" << rule_number << ":;\n";
				}
	;
becomes:  '='			{ $$ = false; }
	| IMPLIES		{ $$ = true; }
	;
lrules:	  lrule			{}
	| lrules lrule		{}
	;
binds:	  exp			{ if (put_gen_lines) *fout << "\n#line " << lineno << '\n';
				  *fout << " NEW_PLANS = merge_plans(NEW_PLANS,"
				       << "LOGICAL" << logical_module_size
				       << "_REWRITE(" << $1 << ",ATRBS,LEVEL+1));\n"; }
	| FAIL			{ *fout << " FOUND = false; goto exit_" << rule_number << "; "; }
	| CASE			{ stopsyms->push('|');
				  *fout << "{ bool FOUND = false;\n Expr EXPR = "; }
	  ccode			{ *fout << ";\n";
				  stopsyms->pop();
				  stopsyms->push(); }
          lcases		{ stopsyms->pop(); }
	  END			{ *fout << "\n if (!FOUND || NEW_PLANS->nullp()) goto exit_" << rule_number << "; }\n";
				  if (put_gen_lines) *fout << "#line " << lineno << '\n'; }
	| FORALL name IN	{ stopsyms->push(DO);
				  *fout << "for(list<Expr>* _" << $2 << " = "; }
	  ccode			{ *fout << "; _" << $2 << "->consp(); _" << $2 << "=_" << $2 << "->tl) { Expr " << $2 << " = _" << $2 << "->hd;\n";
				  stopsyms->pop();
				  stopsyms->push(); }
          DO binds ';'		{ stopsyms->pop(); }
	  END			{ *fout << " };\n";
				  if (put_gen_lines) *fout << "#line " << lineno << '\n'; }
	| LET blist IN binds	{}
	;
bind:	  			{ stopsyms->push(',',IN); }
	  ccode			{ *fout << "; "; stopsyms->pop(); }
	;
blist:	  bind			{}
	| blist ',' bind	{}
	;
lcase:	  '|'			{ cont->open();
				  *fout << "if (!FOUND) {\n"; }
	  cprelude condition IMPLIES binds ';'
				{ *fout << "\n FOUND = true;";
				  if ($4) *fout << " } else FOUND = false;";
				  cont->close();
				  if ($4) *fout << " }";
				  *fout << " };\n"; }
	;
lcases:	  lcase
	| lcases lcase
	;
prule:				{ cont->open(); }
	  patrbs prelude	{ *fout << " if (trace_rules && EXPR->functionp()) trace_input(LEVEL," << rule_number << ",EXPR,ATRBS);\n";
				  in_rule = false;
				  // *fout << " NEW_PLANS = EMPTY_PLANS;\n";
				  rewrite_vars($3,new string("PHYSICAL"));
				  if (put_gen_lines) *fout << "#line " << lineno << '\n';
				  *fout << "{ NEW_PLANS = EMPTY_PLANS;\n";
				  rule_pattern = $3; }
	  condition becomes pbinds ';'
			{ *fout << "if (trace_rules && EXPR->functionp()) trace_output(LEVEL," << rule_number << ",NEW_PLANS);\n";
			  *fout << " PLANS = merge_plans(PLANS,NEW_PLANS);";
			  if ($5) *fout << "\n } else if (trace_rules && EXPR->functionp()) trace(LEVEL," << rule_number << ",\"condition failed\"); };";
			  cont->close();
			  *fout << "} }; exit_" << rule_number << ":;\n";
			}
	;
pbinds:	  exp ':' '{'		{ if (put_gen_lines) *fout << "\n#line " << lineno << '\n';
				  *fout << "synthesized _SA = default_synthesized; "; }
	  synths '}'		{ *fout << "\nNEW_PLANS = NEW_PLANS->cons(new plan(" << $1 << ",_SA));\n";
				  $$=true; }
	| FAIL			{ *fout << " FOUND = false; goto exit_" << rule_number << "; "; }
	| CASE			{ stopsyms->push('|');
				  *fout << "{ bool FOUND = false;\n Expr EXPR = "; }
	  ccode			{ *fout << ";\n";
				  stopsyms->pop();
				  stopsyms->push(); }
          pcases		{ stopsyms->pop(); }
	  END			{ *fout << "\n}\n";
				  if (put_gen_lines) *fout << "#line " << lineno << '\n';
				  $$=$5; }
	| FORALL name IN	{ stopsyms->push(DO);
				  *fout << "for(list<Expr>* _" << $2 << " = "; }
	  ccode			{ *fout << "; _" << $2 << "->consp(); _" << $2 << "=_" << $2 << "->tl) { Expr " << $2 << " = _" << $2 << "->hd;\n";
				  stopsyms->pop();
				  stopsyms->push(); }
          DO pbinds ';'		{ stopsyms->pop(); }
	  END			{ *fout << "\n}"; if (put_gen_lines) *fout << "\n#line " << lineno << '\n';
				  $$=$8; }
	| LET blist IN pbinds	{ $$=$4; }
	;
synth:	  name '='		{ *fout << "_SA." << $1 << " = ";
				  stopsyms->push(';');  }
	  ccode	';'		{ stopsyms->pop(); }
	;
synths:	  synth			{ *fout << "; "; }
	| synths synth		{ *fout << "; "; }
	;
pcase:	  '|'			{ cont->open();
				  *fout << "{\n"; }
	  cprelude condition IMPLIES pbinds ';'
				{ *fout << "\n FOUND = true;";
				  if ($4) *fout << " };";
				  cont->close();
				  if ($4) *fout << " }";
				  *fout << " };\n";
				  $$=$6; }
	;
pcases:	  pcase			{ $$=$1; }
	| pcases pcase		{ if ($1!=$2) yyerror("incompatible cases"); $$=$2; }
	;
patrbs:   rname attributes	{ int lv = 0;
				  in_rule = true;
				  rule_number++;
				  if ($1) *fout << "#define " << $1 << " " << rule_number << '\n';
				  *fout << "{ ";
				  binding<Expr>* atrbs = new binding<Expr>($2->env->reverse());
				  for(Names r = atrbs->names(); r->consp(); r=r->tl)
				     { *fout << "{ ";
				       Expr d = atrbs->find(r->hd);
				       var_decl(d,"Expr");
				       cont->remember();
				       compile_pattern(d,
					   new string("ATRBS." + *r->hd),false);
				     };
				  $$ = lv; }
	;
prules:	  prule
	| prules prule
	;
atrbexp:  name '=' exp ';'	{ $$ = new pair<String,Expr>($1,$3); }
	;
atrbexps: atrbexp		{ $$ = (new binding<Expr>)->extend($1); }
	| atrbexps atrbexp	{ $$ = $1->extend($2); }
	;
attributes:  /* empty */	{ $$ = new binding<Expr>; }
	| '{' atrbpats '}'	{ $$ = $2; }
	;
atrbpat:  name '=' pattern ';'	{ $$ = new pair<String,Expr>($1,$3); }
	;
atrbpats: atrbpat		{ $$ = (new binding<Expr>)->extend($1); }
	| atrbpats atrbpat	{ $$ = $1->extend($2); }
	;
condition: /* empty */		{ $$ = 0; }
	| ':'			{ stopsyms->push('=',IMPLIES);
				  *fout << " { if ("; }
	  ccode			{ stopsyms->pop();
				  *fout << ") { "; $$ = 1; }
	;
rname:	  /* empty */		{ $$ = NULL; }
	| '[' name ']'		{ $$ = $2; }
	;
%%
int main (int argc,char* arg[]) {
  lineno = 1;
  stopsyms->push('#');
  char* ifile = new (GC) char[40];
  String Ifile = new string("");;
  char* strfile = new (GC) char[40];
  strcpy(ifile," ");
  if (argc>1)
  { int len = strlen(arg[1]);
    file_IO = true;
    if (len>4 && arg[1][len-4] == '.' && arg[1][len-3] == 'g'
        && arg[1][len-2] == 'e' && arg[1][len-1] == 'n')
    { strcpy(ifile,arg[1]);
      Ifile = new string(arg[1]);
      yyin = fopen(ifile,"r");
      char* ofile = new (GC) char[len+4];
      strcpy(ofile,arg[1]);
      strcpy(ofile+len-4,"_out.cc");
      fout = new lstream(ofile);
      strfile = new (GC) char[len+3];
      strcpy(strfile,arg[1]);
      strcpy(strfile+len-4,"_str.h");
      fstring.open(strfile);
    } else {
      ifile = new (GC) char[len+4];
      strcpy(ifile,arg[1]);
      strcat(ifile,".gen");
      Ifile = new string(ifile);
      yyin = fopen(ifile,"r");
      char* ofile = new (GC) char[len+7];
      strcpy(ofile,arg[1]);
      strcat(ofile,"_out.cc");
      fout = new lstream(ofile);
      strfile = new (GC) char[len+3];
      strcpy(strfile,arg[1]);
      strcat(strfile,"_str.h");
      fstring.open(strfile);
    };
  *fout << "/***** NOTE: This file was generated automatically by the OPTGEN program ******/\n#include \"external.h\"\n#include \""
        << strfile << "\"\n";
  if (put_gen_lines) *fout << "#line 1 \"" << Ifile << "\"\n";
  yyparse();
  if (file_IO)
     for(list<String>* r = all_strings; r->consp(); r=r->tl)
        fstring << "const Expr _" << r->hd << "_ = variable(\"" << r->hd <<"\");\n";
  fout->flush();
  };
};
