/*********************************************************************************
 *
 *  The server storage manager code. It uses the SHORE SM layer.
 *  Copyright (c) 2002, 2003 by Leonidas Fegaras, the University of Texas at
 *     Arlington. All rights reserved.
 *  Programmer: Leonidas Fegaras
 *  Date: 1/10/02
 *
 ********************************************************************************/

#ifdef LDB_SERVER

#include <stdio.h>
#include <errno.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>

#include "gc_cpp.h"
#include "ldb_sm.h"
#include "server.h"


// 32-bit or 64-bit OIDs
const size_t OID_SIZE = sizeof(serial_t)*8;

const int port_number = 9737; // server port number

extern pinned_object* current_object;  // set when dereferencing an object

extern storage_manager* ldb_sm;  // LDB storage manager

ss_m* ssm; // SHORE storage manager

lvid_t current_volume; // SHORE volume

static int request_socket;

static bool init_device = false;

static const char* device_name;

static long device_size;


static inline int min ( int x, int y ) { return x <? y; };


void ldb_abort ( const char* m ) {
  throw odmg_error_class(m);
};


void sm_error ( const char* file, int line, rc_t rc ) {
  eout << "*** LDB Storage Manager Error (" << file << ":" << line << ")\n" << rc;
  ldb_abort(eout.str());
};


short oid_t::compare ( oid_t x ) const {
  return ((*this)>(serial_t)x) ? 1 : (((*this)<(serial_t)x) ? -1 : 0);
};


pinned_object::pinned_object () {
  tag = PINNED;
  data.po = new pin_i;
};


pinned_object::pinned_object ( pin_i* p ) {
  tag = PINNED;
  data.po = p;
};


oid_t pinned_object::oid () {
  return (tag==PINNED) ? data.po->serial_no() : 0;
};


void pinned_object::repin () {
  data.po->repin();
};


pinned_object::pinned_object ( oid_t oid ) {
  tag = PINNED;
  data.po = new pin_i;
  SM_DO(data.po->pin(current_volume,oid,0,SH));
  pin_cell* c = new pin_cell;
  c->value = data.po;
  c->next = transaction_thread_t::me()->pins;
  transaction_thread_t::me()->pins = c;
};


void clear_pins () {
  for(pin_cell* p = transaction_thread_t::me()->pins; p; )
  {  p->value->unpin();
     delete p->value;
     pin_cell* q = p->next;
     delete p;
     p = q;
  };
  transaction_thread_t::me()->pins = NULL;
};


int pinned_object::body_size () const {
  if (tag==TRANSIENT)
     return data.value.length;
  else return data.po->body_size();
};


sm_string pinned_object::pinned_get ( int offset, size_t size ) {
  static char* s;
  if ( data.po->start_byte() <= offset
       && data.po->start_byte()+data.po->length() >= offset+size )
     return sm_string(data.po->body()+offset-data.po->start_byte(),size);
  else if ( data.po->start_byte() <= offset
            && data.po->start_byte()+data.po->length() > offset )
  {  int i = data.po->start_byte()+data.po->length()-offset;
     s = new (GC) char[size];
     copy_n_bytes(s,data.po->body()+offset-data.po->start_byte(),i);
     while ( i < size )
     {  bool eof;
        SM_DO(data.po->next_bytes(eof));
	if (eof)
	   SM_ERROR(fform("insufficient number of bytes in pinned object: get(%d,%d) from %d bytes",
			 offset,size,body_size()));
        int len = min(data.po->length(),size-i);
        copy_n_bytes(s+i,data.po->body(),len);
	i += len;
     };
     return sm_string(s,size);
  }
  else if (offset+size > body_size())
     SM_ERROR(fform("insufficient number of bytes in pinned object: get(%d,%d) from %d bytes",
		   offset,size,body_size()));
  else {
     SM_DO(data.po->pin(data.po->rid(),offset));
     return sm_string(get(offset,size),size);
  };
};


void pinned_object::pinned_set ( int offset, size_t size, sm_string sdata ) {
  const char* ddata = sdata.content();
  if (offset > body_size())
     SM_ERROR(fform("insufficient number of bytes in pinned object: set(%d,%d) from %d bytes",
		   offset,size,body_size()));
  else if (offset == body_size()) {
     SM_DO(data.po->append_rec(vec_t(ddata,size)));
  } else if (offset+size > body_size()) {
     SM_DO(data.po->update_rec(offset,vec_t(ddata,body_size()-offset)));
     SM_DO(data.po->append_rec(vec_t(ddata+body_size()-offset,size+offset-body_size())));
  } else {
     SM_DO(data.po->update_rec(offset,vec_t(ddata,size)));
  };
};


int pinned_object::pinned_append ( sm_string sdata, size_t data_size ) {
  const char* ddata = sdata.content();
  int no = body_size();
  SM_DO(data.po->append_rec(vec_t(ddata,data_size)));
  return no;
};


/* insert data (of size data_size) after offset+size */
int pinned_object::pinned_append ( int offset, size_t size, sm_string sdata, size_t data_size ) {
  const char* ddata = sdata.content();
  if (offset+size >= body_size()) {
     SM_DO(data.po->append_rec(vec_t(ddata,data_size)));
     return offset;
  } else {
     int no = body_size();
     SM_DO(data.po->append_rec(vec_t(get(offset,size),size)));
     SM_DO(data.po->append_rec(vec_t(ddata,data_size)));
     return no;
  };
};


int pinned_object::pinned_replace ( int offset, size_t size, sm_string sdata, size_t data_size ) {
  const char* ddata = sdata.content();
  if (size >= data_size) {
     set(offset,data_size,ddata);
     return offset;
  } else if (offset+size >= body_size()) {
     set(offset,body_size()-offset,ddata);
     SM_DO(data.po->append_rec(vec_t(ddata+body_size()-offset,data_size-body_size()+offset)));
     return offset;
  } else {
     int no = body_size();
     SM_DO(data.po->append_rec(vec_t(ddata,data_size)));
     return no;
  };
};


sm_string object_deref ( oid_t ref, size_t size ) {
  transaction_thread_t::me()->current_object = new (GC) pinned_object(ref);
  return sm_string(transaction_thread_t::me()->current_object->get(0,size),size);
};


sm_string get_pinned_data ( int offset, size_t size ) {
  return sm_string(transaction_thread_t::me()->current_object->get(offset,size),size);
};


void push_current_object () {
  transaction_thread_t::me()->run_objects[transaction_thread_t::me()->run_object_position++]
        = transaction_thread_t::me()->current_object;
};


void pop_current_object () {
  transaction_thread_t::me()->current_object
        = transaction_thread_t::me()->run_objects[--transaction_thread_t::me()->run_object_position];
};


void set_current_object () {
  transaction_thread_t::me()->current_object
        = transaction_thread_t::me()->run_objects[transaction_thread_t::me()->run_object_position-1];
};


void set_current_object ( oid_t ref ) {
  transaction_thread_t::me()->current_object = new (GC) pinned_object(ref);
};


sm_file::sm_file ( bool temp ) {
  serial_t foid;
  SM_DO(ssm->create_file(current_volume,foid,ss_m::t_regular));
  file_oid = foid;
};


sm_file::sm_file ( oid_t fid ) {
  file_oid=fid;
};


oid_t sm_file::oid () const {
  return file_oid; 
};


sm_ref sm_file::append ( sm_string data ) {
  serial_t oid;
  sm_ref type;
  SM_DO(ssm->create_rec(current_volume,file_oid,vec_t((char*) &type,sizeof(sm_ref)),0,vec_t(data.content(),data.size()),oid));
  transaction_thread_t::me()->current_object = new pinned_object(oid_t(oid));
  return sm_ref(oid_t(oid));
};


void sm_file::remove ( sm_ref object ) {
  SM_DO(ssm->destroy_rec(current_volume,object.oid()));
};


void sm_file::delete_file () {
  SM_DO(ssm->destroy_file(current_volume,file_oid));
};


sm_file_scan::sm_file_scan ( sm_file file )
  : oid(new scan_file_i(current_volume,file.oid()))
  {};


void sm_file_scan::close () {
  oid->finish();
  delete oid;
};


bool sm_file_scan::next ( sm_ref &object ) {
  pin_i* po;
  bool eof;
  SM_DO(oid->next(po,0,eof));
  if (!eof)
     object = sm_ref(po->serial_no());
  return !eof;
};


bool sm_file_scan::eof () {
  return oid->eof();
};


sm_file_append::sm_file_append ( sm_file file )
  : oid(new append_file_i(current_volume,file.oid()))
  {};


void sm_file_append::append ( sm_string data ) {
  lrid_t loid;
  sm_ref type;
  SM_DO(oid->create_rec(vec_t((char*) &type,sizeof(sm_ref)),0,vec_t(data.content(),data.size()),loid));
};


void sm_file_append::close () {
  oid->finish();
  delete oid;
};


sm_index::sm_index ( bool temp, bool uniquep, sm_string kdesc ) {
  serial_t soid;
  unique_index = uniquep;
  SM_DO(ssm->create_index(current_volume,
			  (uniquep) ? ss_m::t_uni_btree : ss_m::t_btree,
			  ss_m::t_regular,kdesc.content(),0,soid));
  index_oid = soid;
};


sm_index::sm_index ( oid_t iid, bool uniquep ) {
  unique_index = uniquep;
  index_oid=iid;
};


oid_t sm_index::oid () const {
  return index_oid;
};


void sm_index::add ( sm_string key, raw_ref data ) {
  if (unique_index && member(key))
     SM_ERROR(fform("duplicate key in unique index: %s",key.scontent()));
  SM_DO(ssm->create_assoc(current_volume,index_oid,
			  vec_t(key.content(),key.size()),
			  vec_t((char*) &data,sizeof(raw_ref))));
};


void sm_index::remove ( sm_string key, raw_ref data ) {
  SM_DO(ssm->destroy_assoc(current_volume,index_oid,
			   vec_t(key.content(),key.size()),
			   vec_t((char*) &data,sizeof(raw_ref))));
};


bool sm_index::member ( sm_string key ) {
  bool found;
  static char* dv = new (GC) char[sizeof(raw_ref)];
  smsize_t len = sizeof(raw_ref);
  SM_DO(ssm->find_assoc(current_volume,index_oid,
			vec_t(key.content(),key.size()),
			(void*) dv,len,found));
  return found;
};


bool sm_index::find ( sm_string key, raw_ref &data ) {
  bool found;
  smsize_t len = sizeof(raw_ref);
  SM_DO(ssm->find_assoc(current_volume,index_oid,
			vec_t(key.content(),key.size()),
			(void*) &data,len,found));
  return found;
};


void sm_index::delete_index () {
  SM_DO(ssm->destroy_index(current_volume,index_oid));
};


sm_index_scan::sm_index_scan ( sm_index index, sm_string low, sm_string high )
  : oid(new scan_index_i(current_volume,index.oid(),
			 scan_index_i::ge,cvec_t(low.content(),low.size()),
			 scan_index_i::le,cvec_t(high.content(),high.size())))
  {};


void sm_index_scan::close () {
  oid->finish();
  delete oid;
};


bool sm_index_scan::next ( sm_string &key, raw_ref &data ) {
  bool eof;
  SM_DO(oid->next(eof));
  if (!eof) {
     static char* ikey = new (GC) char[1000];
     smsize_t klen = 1000;
     smsize_t elen = sizeof(raw_ref);
     SM_DO(oid->curr(new vec_t(ikey,klen),klen,new vec_t((char*) &data,elen),elen));
     key = sm_string(ikey,klen);
  };
  return !eof;
};


bool sm_index_scan::eof () {
  return oid->eof(); 
};


key_info_t* sort_key () {
  key_info_t* k = new (GC) key_info_t();
  k->type = sortorder::kt_b;
  k->len = 100;
  return k;
};


sort_parm_t* sort_parameters ( bool uniquep ) {
  sort_parm_t* s = new (GC) sort_parm_t();
  s->unique = uniquep;            // result unique ?
  s->run_size = 100;              // size for each run (# of pages)
  s->ascending = true;            // ascending order ?
  s->lvid = current_volume;       // logical volume id
  s->property = ss_m::t_regular;  // temporary file ?
  return s;
};


sm_sort_scan::sm_sort_scan  ( bool uniquep )
  : oid(new sort_stream_i(*sort_key(),*sort_parameters(uniquep)))
  {};


void sm_sort_scan::close () {
  oid->finish();
  delete oid;
};


void sm_sort_scan::put ( sm_string key, sm_string data ) {
  SM_DO(oid->put(vec_t(key.content(),key.size()),vec_t(data.content(),data.size())));
};


bool sm_sort_scan::next ( sm_string &key, sm_string &data ) {
  vec_t vkey, vdata;
  bool eof;
  SM_DO(oid->get_next(vkey,vdata,eof));
  key = sm_string((const char*)vkey.ptr(0),vkey.len(0));
  data = sm_string((const char*)vdata.ptr(0),vdata.len(0));
  return eof;
};


bool sm_sort_scan::eof () {
  return oid->is_empty();
};


void storage_manager::begin_transaction () {
  if (in_transaction)
     SM_ERROR("Already in a transaction");
  in_transaction = true;
  SM_DO(ssm->begin_xct());
  serial_t c;
  SM_DO(ssm->vol_root_index(current_volume,c));
  catalog = oid_t(c);
};


void storage_manager::commit_transaction () {
  if (!in_transaction)
     SM_ERROR("Not in a transaction");
  in_transaction = false; 
  clear_pins();
  catalog = oid_t();
  SM_DO(ssm->commit_xct());
};


void storage_manager::abort_transaction () {
  if (!in_transaction)
     SM_ERROR("Not in a transaction");
  in_transaction = false; 
  clear_pins();
  SM_DO(ssm->abort_xct());
};


void storage_manager::add_declaration ( sm_string name, sm_string value ) {
  static char* el = new (GC) char;
  smsize_t elen = 0;
  bool found;
  SM_DO(ssm->find_assoc(current_volume,catalog,
			vec_t(name.content(),name.size()),
			el,elen,found));
  if (found)
     SM_ERROR(fform("Declaration '%s' already exists",name.content()));
  SM_DO(ssm->create_assoc(current_volume,catalog,vec_t(name.content(),name.size()),
			  vec_t(value.content(),value.size())));
};


bool storage_manager::find_declaration ( sm_string name, sm_string &value ) {
  static char* el = new (GC) char[10000];
  smsize_t elen = 10000;
  bool found;
  SM_DO(ssm->find_assoc(current_volume,catalog,
			vec_t(name.content(),name.size()),
			el,elen,found));
  if (found)
     value = sm_string(el,elen);
  return found;
};


void storage_manager::remove_declaration ( sm_string name ) {
  int num;
  SM_DO(ssm->destroy_all_assoc(current_volume,catalog,
			       vec_t(name.content(),name.size()),
			       num));
};


rc_t init_config_options ( const char* opt_file, option_group_t& options ) {
  rc_t rc;	// return code
  /* have the SSM add its options to the group */
  W_DO(ss_m::setup_options(&options));
  /* read the config file to set options */
  ostrstream err_stream;
  option_file_scan_t opt_scan(opt_file,&options);
  /* scan the file and override any current option settings */
  /* options names must be spelled correctly   */
  rc = opt_scan.scan(true,err_stream,true);
  if (rc)
  {  char* errmsg = err_stream.str();
     cerr << "Error in reading option file: " << opt_file << endl << "\t" << errmsg << endl;
     return rc;
  };
  /* check required options */
  rc = options.check_required(&err_stream);
  char* errmsg = err_stream.str();
  if (rc)
  {  cerr << "These required options are not set:" << endl << errmsg << endl;
     return rc;
  };
  return RCOK;
};


rc_t init_device_and_volume ( lvid_t& lvid ) {
  u_int vol_cnt;
  devid_t devid;
  if (init_device)
  {  cout << "Creating and formatting the device: " << device_name << endl;
     cout << "Size = " << device_size << " Kbytes\n";
     W_DO(ssm->format_dev(device_name,device_size,true));
     W_DO(ssm->mount_dev(device_name,vol_cnt,devid));
     W_DO(ssm->generate_new_lvid(lvid));
     W_DO(ssm->create_vol(device_name,lvid,device_size));
     W_DO(ssm->add_logical_id_index(lvid,0,0));
  } else {
     W_DO(ssm->mount_dev(device_name,vol_cnt,devid));
     lvid_t* lvid_list;
     u_int   lvid_cnt;
     W_DO(ssm->list_volumes(device_name, lvid_list, lvid_cnt));
     if (lvid_cnt == 0) {
	cerr << "Device has no volumes" << endl;
	::exit(1);
     };
     lvid = lvid_list[0];
     delete [] lvid_list;
  };
  return RCOK;
};


startup_smthread_t::startup_smthread_t ()
 : smthread_t(t_regular, "startup", WAIT_FOREVER, sthread_t::default_stack * 100)
{};


transaction_smthread_t::transaction_smthread_t ()
 : smthread_t(t_regular, "transaction", WAIT_FOREVER, sthread_t::default_stack * 100)
{};


static void (*storage_process)();


void transaction_smthread_t::run () {
  (*storage_process)();
};


void do_transaction () {
  transaction_smthread_t *doit = new transaction_smthread_t();
  if (!doit) W_FATAL(fcOUTOFMEMORY);
  SM_DO(doit->fork());
  SM_DO(doit->wait());
  delete doit;
};


void startup_smthread_t::run () {
  ssm = new ss_m();
  if (!ssm) {
     cerr << "Error: Out of memory for the storage manager" << endl;
     return;
  };
  SM_DO(init_device_and_volume(current_volume));
  do_transaction();
  delete ssm;
};


storage_manager::storage_manager ( sm_string device_pathname, bool format_device, long size ) {
  init_device = format_device;
  device_name = device_pathname.content();
  device_size = size;
  in_transaction = false; 
};


void storage_manager::run ( void (*program) () ) {
  storage_process = program;
  startup_smthread_t *doit = new startup_smthread_t();
  if (!doit)
     W_FATAL(fcOUTOFMEMORY);
  SM_DO(doit->fork());
  SM_DO(doit->wait());
  delete doit;
};


/*****************************************************************************************************/


cleanup_thread_t::cleanup_thread_t ()
 : smthread_t(t_regular, "cleanup", WAIT_FOREVER, sthread_t::default_stack * 2)
 {};


static list<transaction_thread_t*>* client_list;

static smutex_t* client_list_mutex = new smutex_t("client list mutex");

void cleanup_thread_t::run () {
  while (true)
  {  SM_DO(client_list_mutex->acquire());
     list<transaction_thread_t*>* s = client_list;
     if (client_list->consp())
     {  for (list<transaction_thread_t*>* r = client_list->tl; r->consp(); r=r->tl, s=s->tl)
	     if (!r->hd->active)
	     {  delete r->hd;
	        s->tl = r->tl;
	     };
        if (!client_list->hd->active)
	{  delete client_list->hd;
           client_list = client_list->tl;
	};
     };
     if (false && client_list->nullp())
       cout << "No more clients\n";
     client_list_mutex->release();
     yield();
  };
};


startup_thread_t::startup_thread_t ()
 : smthread_t(t_regular, "startup", WAIT_FOREVER, sthread_t::default_stack * 2)
 {};


transaction_thread_t::transaction_thread_t ()
  : smthread_t(t_regular, "transaction", WAIT_FOREVER, sthread_t::default_stack * 100) {
  /* Create the socket and set it up to accept connections. */
#if defined(linux) || defined(__linux__)
  static size_t size = sizeof(clientname);
#else
  static int size = sizeof(clientname);
#endif
  sock = accept(request_socket,(struct sockaddr *) &clientname,&size);
  if (sock < 0)
     SM_ERROR("Cannot accept a new socket connection");
  cout << "Connecting from host " << inet_ntoa(clientname.sin_addr)
       << ", port " << ntohs(clientname.sin_port) << ".\n";
  stream = rpc_stream(sock);
  pins  = NULL;
  run_object_position = 0;
};


void transaction_thread_t::run () {
  fd_set active_fd_set;
  fd_set read_fd_set;
  FD_ZERO(&active_fd_set);
  FD_SET(sock,&active_fd_set);
  while (true)
  {   /* Block until input arrives on the client socket. */
      read_fd_set = active_fd_set;
      if (select(FD_SETSIZE,&read_fd_set,NULL,NULL,NULL) < 0)
	 SM_ERROR("Cannot select active sockets");
      /* Service the client socket if its input is pending . */
      if (FD_ISSET(sock,&read_fd_set))
      {  short command;
	 try {
	   stream >> command;
	 } catch (odmg_error_class(m)) {
	   close(sock);
	   FD_CLR(sock,&active_fd_set);
	   cout << "Closing connection from host "
	        << inet_ntoa(clientname.sin_addr) << ", port "
	        << ntohs(clientname.sin_port) << ".\n";
	   return;
	 };
	 if (command<0 || command>RPC_fnc_count)
	 {  close(sock);
	    FD_CLR(sock,&active_fd_set);
	    cout << "Closing connection from host "
	         << inet_ntoa(clientname.sin_addr) << ", port "
	         << ntohs(clientname.sin_port) << ".\n";
	    return;
	 } else {
	   try {
	     /* Dispatch client call */
	     (*dispatch_table[command])(stream);
	   } catch (odmg_error_class(m)) {
	     char res = 1;
	     send(sock,&res,1,0);
	     stream << m.message;
	     eout.reset();
	   };
	 };
      };
  };
};


int make_socket ( uint16_t port ) {
  int sock;
  struct sockaddr_in name;
  /* Create the socket. */
  sock = socket(PF_INET,SOCK_STREAM,0);
  if (sock < 0)
     SM_ERROR("Cannot create socket");
  /* Give the socket a name. */
  name.sin_family = AF_INET;
  name.sin_port = htons(port);
  name.sin_addr.s_addr = htonl(INADDR_ANY);
  if (bind(sock,(struct sockaddr *) &name,sizeof(name)) < 0)
     SM_ERROR("Cannot bind socket");
  return sock;
};


void startup_thread_t::run () {
  fd_set active_fd_set;
  fd_set read_fd_set;
  struct timeval tv;
  ssm = new ss_m();
  if (!ssm) {
     cerr << "Error: Out of memory for the storage manager" << endl;
     return;
  };
  SM_DO(init_device_and_volume(current_volume));
  /* Create the socket and set it up to accept connections. */
  for(int i = port_number; listen(request_socket,1)<0 ; i++)
     request_socket = make_socket(i);
  /* Initialize the set of active sockets. */
  FD_ZERO(&active_fd_set);
  FD_SET(request_socket,&active_fd_set);
  cout << "The LDB server is now running...\n";
  while (true)
  {  /* Block until input arrives on the listener socket. */
     read_fd_set = active_fd_set;
     tv.tv_sec = 0;
     tv.tv_usec = 1;
     if (select(FD_SETSIZE,&read_fd_set,NULL,NULL,&tv) < 0)
	SM_ERROR("Cannot select active socket");
     if (FD_ISSET(request_socket,&read_fd_set))
     {  /* request for new client */
	transaction_thread_t *doit = new transaction_thread_t();
	if (!doit)
	   W_FATAL(fcOUTOFMEMORY);
	SM_DO(doit->fork());
        SM_DO(doit->wait());
	delete doit;
     };
     this->yield();
  };
};


void storage_manager::run () {
  startup_thread_t *doit = new startup_thread_t();
  if (!doit)
     W_FATAL(fcOUTOFMEMORY);
  try { SM_DO(doit->fork());
        SM_DO(doit->wait());
  } catch (odmg_error_class(m)) {
    cerr << m.message;
    w_base_t::abort();
  };
  delete doit;
};

#endif
